#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from datetime import date, time, datetime, timedelta
from decimal import Decimal
from typing import Any, Optional, List
from unittest import TestCase

from isodate import Duration

from jdmn.feel.lib.BaseStandardFEELLib import BaseStandardFEELLib
from jdmn.feel.lib.DefaultStandardFEELLib import DefaultStandardFEELLib
from jdmn.runtime.Assert import Assert
from jdmn.runtime.Context import Context
from jdmn.runtime.Range import Range


class FEELOperatorsTest(TestCase):
    """
    Base test class for operators
    """
    __test__ = False

    #
    # Numeric operators
    #
    def testIsNumeric(self):
        self.assertFalse(self.getLib().isNumber(None))
        self.assertTrue(self.getLib().isNumber(self.getLib().number("1")))
        self.assertFalse(self.getLib().isNumber("abc"))
        self.assertFalse(self.getLib().isNumber(True))
        self.assertFalse(self.getLib().isNumber(self.getLib().date("2020-01-01")))
        self.assertFalse(self.getLib().isNumber(self.getLib().time("12:00:00")))
        self.assertFalse(self.getLib().isNumber(self.getLib().dateAndTime("2020-01-01T12:00:00")))
        self.assertFalse(self.getLib().isNumber(self.getLib().duration("P1Y1M")))
        self.assertFalse(self.getLib().isNumber(self.getLib().asList("a")))
        self.assertFalse(self.getLib().isNumber(Context()))
        self.assertFalse(self.getLib().isNumber(Range(True, 0, True, 1)))

    def testNumericValue(self):
        self.assertIsNone(self.getLib().numericValue(None))

        self.assertEqual(self.makeNumber("1"), self.getLib().numericValue(self.makeNumber("1")))

    def testNumericIs(self):
        self.assertTrue(self.getLib().numericIs(None, None))
        self.assertFalse(self.getLib().numericIs(None, self.makeNumber("1")))
        self.assertFalse(self.getLib().numericIs(self.makeNumber("1"), None))

        self.assertTrue(self.getLib().numericIs(self.makeNumber("1"), self.makeNumber("1")))
        self.assertFalse(self.getLib().numericIs(self.makeNumber("1"), self.makeNumber("2")))

    def testNumericEqual(self):
        self.assertTrue(self.getLib().numericEqual(None, None))
        self.assertFalse(self.getLib().numericEqual(None, self.makeNumber("1")))
        self.assertFalse(self.getLib().numericEqual(self.makeNumber("1"), None))

        self.assertTrue(self.getLib().numericEqual(self.makeNumber("1"), self.makeNumber("1")))
        self.assertFalse(self.getLib().numericEqual(self.makeNumber("1"), self.makeNumber("2")))

    def testNumericNotEqual(self):
        self.assertFalse(self.getLib().numericNotEqual(None, None))
        self.assertTrue(self.getLib().numericNotEqual(None, self.makeNumber("1")))
        self.assertTrue(self.getLib().numericNotEqual(self.makeNumber("1"), None))

        self.assertFalse(self.getLib().numericNotEqual(self.makeNumber("1"), self.makeNumber("1")))
        self.assertTrue(self.getLib().numericNotEqual(self.makeNumber("1"), self.makeNumber("2")))

    def testNumericLessThan(self):
        self.assertIsNone(self.getLib().numericLessThan(None, None))
        self.assertIsNone(self.getLib().numericLessThan(None, self.makeNumber("1")))
        self.assertIsNone(self.getLib().numericLessThan(self.makeNumber("1"), None))

        self.assertFalse(self.getLib().numericLessThan(self.makeNumber("1"), self.makeNumber("1")))
        self.assertTrue(self.getLib().numericLessThan(self.makeNumber("1"), self.makeNumber("2")))

    def testNumericGreaterThan(self):
        self.assertIsNone(self.getLib().numericGreaterThan(None, None))
        self.assertIsNone(self.getLib().numericGreaterThan(None, self.makeNumber("1")))
        self.assertIsNone(self.getLib().numericGreaterThan(self.makeNumber("1"), None))

        self.assertFalse(self.getLib().numericGreaterThan(self.makeNumber("1"), self.makeNumber("1")))
        self.assertTrue(self.getLib().numericGreaterThan(self.makeNumber("2"), self.makeNumber("1")))

    def testNumericLessEqualThan(self):
        self.assertTrue(self.getLib().numericLessEqualThan(None, None))
        self.assertIsNone(self.getLib().numericLessEqualThan(None, self.makeNumber("1")))
        self.assertIsNone(self.getLib().numericLessEqualThan(self.makeNumber("1"), None))

        self.assertTrue(self.getLib().numericLessEqualThan(self.makeNumber("1"), self.makeNumber("1")))
        self.assertFalse(self.getLib().numericLessEqualThan(self.makeNumber("2"), self.makeNumber("1")))

    def testNumericGreaterEqualThan(self):
        self.assertTrue(self.getLib().numericGreaterEqualThan(None, None))
        self.assertIsNone(self.getLib().numericGreaterEqualThan(None, self.makeNumber("1")))
        self.assertIsNone(self.getLib().numericGreaterEqualThan(self.makeNumber("1"), None))

        self.assertTrue(self.getLib().numericGreaterEqualThan(self.makeNumber("1"), self.makeNumber("1")))
        self.assertFalse(self.getLib().numericGreaterEqualThan(self.makeNumber("1"), self.makeNumber("2")))

    def testNumericAdd(self):
        self.assertIsNone(self.getLib().numericAdd(None, None))
        self.assertIsNone(self.getLib().numericAdd(None, self.makeNumber("1")))
        self.assertIsNone(self.getLib().numericAdd(self.makeNumber("1"), None))

        self.assertEqualsNumber(self.makeNumber("3"), self.getLib().numericAdd(self.makeNumber("1"), self.makeNumber("2")))

    def testNumericSubtract(self):
        self.assertIsNone(self.getLib().numericSubtract(None, None))
        self.assertIsNone(self.getLib().numericSubtract(None, self.makeNumber("2")))
        self.assertIsNone(self.getLib().numericSubtract(self.makeNumber("2"), None))

        self.assertEqualsNumber(self.makeNumber("-1"), self.getLib().numericSubtract(self.makeNumber("1"), self.makeNumber("2")))

    def testNumericMultiply(self):
        self.assertIsNone(self.getLib().numericMultiply(None, None))
        self.assertIsNone(self.getLib().numericMultiply(None, self.makeNumber("2")))
        self.assertIsNone(self.getLib().numericMultiply(self.makeNumber("2"), None))

        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().numericMultiply(self.makeNumber("1"), self.makeNumber("2")))

    def testNumericDivide(self):
        self.assertIsNone(self.getLib().numericDivide(None, None))
        self.assertIsNone(self.getLib().numericDivide(None, self.makeNumber("2")))
        self.assertIsNone(self.getLib().numericDivide(self.makeNumber("2"), None))

        self.assertEqualsNumber("0.5", self.getLib().numericDivide(self.makeNumber("1"), self.makeNumber("2")))

    def testNumericUnaryMinus(self):
        self.assertIsNone(self.getLib().numericUnaryMinus(None))

        self.assertEqualsNumber(self.makeNumber("-1"), self.getLib().numericUnaryMinus(self.makeNumber("1")))

    def testNumericExponentiation(self):
        self.assertIsNone(self.getLib().numericExponentiation(None, None))
        self.assertIsNone(self.getLib().numericExponentiation(None, self.makeNumber("10")))
        self.assertIsNone(self.getLib().numericExponentiation(self.makeNumber("10"), None))

        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().numericExponentiation(self.makeNumber("2"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("0.5"), self.getLib().numericExponentiation(self.makeNumber("2"), self.makeNumber("-1")))
        self.assertEqualsNumber(self.makeNumber("1024"), self.getLib().numericExponentiation(self.makeNumber("2"), self.makeNumber("10")))

        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().numericExponentiation(self.makeNumber("3"), self.makeNumber("0")))

        self.assertEqualsNumber(self.makeNumber("1.41421356"), self.getLib().numericExponentiation(self.makeNumber("2"), self.makeNumber("0.5")))
        self.assertEqualsNumber(self.makeNumber("8"), self.getLib().numericExponentiation(self.makeNumber("2"), self.makeNumber("3")))
        self.assertEqualsNumber(self.makeNumber("11.31370849"), self.getLib().numericExponentiation(self.makeNumber("2"), self.makeNumber("3.5")))
        self.assertEqualsNumber(self.makeNumber("11.84466611"), self.getLib().numericExponentiation(self.makeNumber("3"), self.makeNumber("2.25")))
        self.assertEqualsNumber(self.makeNumber("15.58845726"), self.getLib().numericExponentiation(self.makeNumber("3"), self.makeNumber("2.5")))
        self.assertEqualsNumber(self.makeNumber("20.51556351"), self.getLib().numericExponentiation(self.makeNumber("3"), self.makeNumber("2.75")))
        self.assertEqualsNumber(self.makeNumber("1.73205080"), self.getLib().numericExponentiation(self.makeNumber("3"), self.makeNumber("0.5")))
        self.assertEqualsNumber(self.makeNumber("0.11111111"), self.getLib().numericExponentiation(self.makeNumber("3"), self.makeNumber("-2")))
        self.assertEqualsNumber(self.makeNumber("0.04874348"), self.getLib().numericExponentiation(self.makeNumber("3"), self.makeNumber("-2.75")))

    #
    # Boolean operators
    #
    def testIsBoolean(self):
        self.assertFalse(self.getLib().isBoolean(None))
        self.assertFalse(self.getLib().isBoolean(self.getLib().number("1")))
        self.assertFalse(self.getLib().isBoolean("abc"))
        self.assertTrue(self.getLib().isBoolean(True))
        self.assertFalse(self.getLib().isBoolean(self.getLib().date("2020-01-01")))
        self.assertFalse(self.getLib().isBoolean(self.getLib().time("12:00:00")))
        self.assertFalse(self.getLib().isBoolean(self.getLib().dateAndTime("2020-01-01T12:00:00")))
        self.assertFalse(self.getLib().isBoolean(self.getLib().duration("P1Y1M")))
        self.assertFalse(self.getLib().isBoolean(self.getLib().asList("a")))
        self.assertFalse(self.getLib().isBoolean(Context()))
        self.assertFalse(self.getLib().isBoolean(Range(True, 0, True, 1)))

    def testBooleanValue(self):
        self.assertIsNone(self.getLib().booleanValue(None))
        self.assertTrue(self.getLib().booleanValue(True))
        self.assertFalse(self.getLib().booleanValue(False))

    def testBooleanIs(self):
        self.assertTrue(self.getLib().booleanIs(None, None))
        self.assertFalse(self.getLib().booleanIs(True, None))
        self.assertFalse(self.getLib().booleanIs(None, True))

        self.assertFalse(self.getLib().booleanIs(False, True))
        self.assertTrue(self.getLib().booleanIs(True, True))

    def testBooleanEqual(self):
        self.assertTrue(self.getLib().booleanEqual(None, None))
        self.assertFalse(self.getLib().booleanEqual(True, None))
        self.assertFalse(self.getLib().booleanEqual(None, True))

        self.assertFalse(self.getLib().booleanEqual(False, True))
        self.assertTrue(self.getLib().booleanEqual(True, True))

    def testBooleanNotEqual(self):
        self.assertFalse(self.getLib().booleanNotEqual(None, None))
        self.assertTrue(self.getLib().booleanNotEqual(True, None))
        self.assertTrue(self.getLib().booleanNotEqual(None, True))

        self.assertTrue(self.getLib().booleanNotEqual(False, True))
        self.assertFalse(self.getLib().booleanNotEqual(True, True))

    def testBooleanNot(self):
        self.assertTrue(self.getLib().booleanNot(False))
        self.assertFalse(self.getLib().booleanNot(True))
        self.assertIsNone(self.getLib().booleanNot(None))

        self.assertIsNone(self.getLib().booleanNot("abc"))
        self.assertIsNone(self.getLib().booleanNot(self.makeNumber("123")))

    def testBooleanOr(self):
        self.assertFalse(self.getLib().booleanOr(False, False))
        self.assertTrue(self.getLib().booleanOr(False, True))
        self.assertTrue(self.getLib().booleanOr(True, False))
        self.assertTrue(self.getLib().booleanOr(True, True))
        self.assertIsNone(self.getLib().booleanOr(False, None))
        self.assertIsNone(self.getLib().booleanOr(None, False))
        self.assertTrue(self.getLib().booleanOr(True, None))
        self.assertTrue(self.getLib().booleanOr(None, True))
        self.assertTrue(self.getLib().booleanOr(True, self.makeNumber("123")))
        self.assertTrue(self.getLib().booleanOr(self.makeNumber("123"), True))
        self.assertTrue(self.getLib().booleanOr(True, "123"))
        self.assertTrue(self.getLib().booleanOr(True, None))
        self.assertTrue(self.getLib().booleanOr(None, True))
        self.assertTrue(self.getLib().booleanOr("123", True))
        self.assertIsNone(self.getLib().booleanOr(None, None))
        self.assertIsNone(self.getLib().booleanOr("123", "123"))

        self.assertIsNone(self.getLib().booleanOr(True))
        self.assertTrue(self.getLib().booleanOr(False, False, True))
        self.assertTrue(self.getLib().booleanOr([False, False, True]))
        self.assertIsNone(self.getLib().booleanOr([[False, False, True]]))

    def testBinaryBooleanOr(self):
        self.assertFalse(self.getLib().binaryBooleanOr(False, False))
        self.assertTrue(self.getLib().binaryBooleanOr(False, True))
        self.assertTrue(self.getLib().binaryBooleanOr(True, False))
        self.assertTrue(self.getLib().binaryBooleanOr(True, True))
        self.assertIsNone(self.getLib().binaryBooleanOr(False, None))
        self.assertIsNone(self.getLib().binaryBooleanOr(None, False))
        self.assertTrue(self.getLib().binaryBooleanOr(True, None))
        self.assertTrue(self.getLib().binaryBooleanOr(None, True))
        self.assertTrue(self.getLib().binaryBooleanOr(True, self.makeNumber("123")))
        self.assertTrue(self.getLib().binaryBooleanOr(self.makeNumber("123"), True))
        self.assertTrue(self.getLib().binaryBooleanOr(True, "123"))
        self.assertTrue(self.getLib().binaryBooleanOr(True, None))
        self.assertTrue(self.getLib().binaryBooleanOr(None, True))
        self.assertTrue(self.getLib().binaryBooleanOr("123", True))
        self.assertIsNone(self.getLib().binaryBooleanOr(None, None))
        self.assertIsNone(self.getLib().binaryBooleanOr("123", "123"))

    def testBooleanAnd(self):
        self.assertFalse(self.getLib().booleanAnd(False, False))
        self.assertFalse(self.getLib().booleanAnd(False, True))
        self.assertFalse(self.getLib().booleanAnd(True, False))
        self.assertTrue(self.getLib().booleanAnd(True, True))
        self.assertFalse(self.getLib().booleanAnd(False, None))
        self.assertFalse(self.getLib().booleanAnd(None, False))
        self.assertFalse(self.getLib().booleanAnd(False, self.makeNumber("123")))
        self.assertFalse(self.getLib().booleanAnd(self.makeNumber("123"), False))
        self.assertFalse(self.getLib().booleanAnd(False, "123"))
        self.assertFalse(self.getLib().booleanAnd(False, None))
        self.assertFalse(self.getLib().booleanAnd(None, False))
        self.assertFalse(self.getLib().booleanAnd("123", False))
        self.assertIsNone(self.getLib().booleanAnd(True, None))
        self.assertIsNone(self.getLib().booleanAnd(None, True))
        self.assertIsNone(self.getLib().booleanAnd(None, None))
        self.assertIsNone(self.getLib().booleanAnd("123", "123"))

        self.assertIsNone(self.getLib().booleanAnd(True))
        self.assertFalse(self.getLib().booleanAnd(True, True, False))
        self.assertFalse(self.getLib().booleanAnd([False, False, True]))
        self.assertIsNone(self.getLib().booleanAnd([[False, False, True]]))

    def testBinaryBooleanAnd(self):
        self.assertFalse(self.getLib().binaryBooleanAnd(False, False))
        self.assertFalse(self.getLib().binaryBooleanAnd(False, True))
        self.assertFalse(self.getLib().binaryBooleanAnd(True, False))
        self.assertTrue(self.getLib().binaryBooleanAnd(True, True))
        self.assertFalse(self.getLib().binaryBooleanAnd(False, None))
        self.assertFalse(self.getLib().binaryBooleanAnd(None, False))
        self.assertFalse(self.getLib().binaryBooleanAnd(False, self.makeNumber("123")))
        self.assertFalse(self.getLib().binaryBooleanAnd(self.makeNumber("123"), False))
        self.assertFalse(self.getLib().binaryBooleanAnd(False, "123"))
        self.assertFalse(self.getLib().binaryBooleanAnd(False, None))
        self.assertFalse(self.getLib().binaryBooleanAnd(None, False))
        self.assertFalse(self.getLib().binaryBooleanAnd("123", False))
        self.assertIsNone(self.getLib().binaryBooleanAnd(True, None))
        self.assertIsNone(self.getLib().binaryBooleanAnd(None, True))
        self.assertIsNone(self.getLib().binaryBooleanAnd(None, None))
        self.assertIsNone(self.getLib().binaryBooleanAnd("123", "123"))

    #
    # String operators
    #
    def testIsString(self):
        self.assertFalse(self.getLib().isString(None))
        self.assertFalse(self.getLib().isString(self.getLib().number("1")))
        self.assertTrue(self.getLib().isString("abc"))
        self.assertFalse(self.getLib().isString(True))
        self.assertFalse(self.getLib().isString(self.getLib().date("2020-01-01")))
        self.assertFalse(self.getLib().isString(self.getLib().time("12:00:00")))
        self.assertFalse(self.getLib().isString(self.getLib().dateAndTime("2020-01-01T12:00:00")))
        self.assertFalse(self.getLib().isString(self.getLib().duration("P1Y1M")))
        self.assertFalse(self.getLib().isString(self.getLib().asList("a")))
        self.assertFalse(self.getLib().isString(Context()))
        self.assertFalse(self.getLib().isString(Range(True, 0, True, 1)))

    def testStringValue(self):
        self.assertIsNone(self.getLib().stringValue(None))
        self.assertEqual("a", self.getLib().stringValue("a"))

    def testStringIs(self):
        self.assertTrue(self.getLib().stringIs(None, None))
        self.assertFalse(self.getLib().stringIs("a", None))
        self.assertFalse(self.getLib().stringIs(None, "b"))

        self.assertFalse(self.getLib().stringIs("a", "b"))
        self.assertTrue(self.getLib().stringIs("b", "b"))

    def testStringEqual(self):
        self.assertTrue(self.getLib().stringEqual(None, None))
        self.assertFalse(self.getLib().stringEqual("a", None))
        self.assertFalse(self.getLib().stringEqual(None, "b"))

        self.assertFalse(self.getLib().stringEqual("a", "b"))
        self.assertTrue(self.getLib().stringEqual("b", "b"))

    def testStringNotEqual(self):
        self.assertFalse(self.getLib().stringNotEqual(None, None))
        self.assertTrue(self.getLib().stringNotEqual("a", None))
        self.assertTrue(self.getLib().stringNotEqual(None, "b"))

        self.assertTrue(self.getLib().stringNotEqual("a", "b"))
        self.assertFalse(self.getLib().stringNotEqual("b", "b"))

    def testStringAdd(self):
        self.assertIsNone(self.getLib().stringAdd(None, None))
        self.assertIsNone(self.getLib().stringAdd("a", None))
        self.assertIsNone(self.getLib().stringAdd(None, "b"))

        self.assertEqual("ab", self.getLib().stringAdd("a", "b"))
        self.assertEqual("ba", self.getLib().stringAdd("b", "a"))

    #
    # Date operators
    #
    def testIsDate(self):
        self.assertFalse(self.getLib().isDate(None))
        self.assertFalse(self.getLib().isDate(self.getLib().number("1")))
        self.assertFalse(self.getLib().isDate("abc"))
        self.assertFalse(self.getLib().isDate(True))
        self.assertTrue(self.getLib().isDate(self.getLib().date("2020-01-01")))
        self.assertFalse(self.getLib().isDate(self.getLib().time("12:00:00")))
        self.assertFalse(self.getLib().isDate(self.getLib().dateAndTime("2020-01-01T12:00:00")))
        self.assertFalse(self.getLib().isDate(self.getLib().duration("P1Y1M")))
        self.assertFalse(self.getLib().isDate(self.getLib().asList("a")))
        self.assertFalse(self.getLib().isDate(Context()))
        self.assertFalse(self.getLib().isDate(Range(True, 0, True, 1)))

    def testDateValue(self):
        self.assertIsNone(self.getLib().dateValue(None))

        self.assertEqual(0, self.getLib().dateValue(self.makeDate("1970-01-01")))
        self.assertEqual(86400, self.getLib().dateValue(self.makeDate("1970-01-02")))
        self.assertEqual(2678400, self.getLib().dateValue(self.makeDate("1970-02-01")))
        self.assertEqual(946684800, self.getLib().dateValue(self.makeDate("2000-01-01")))

        # Negative dates are not supported in datetime object
        # self.assertEqual(-124334265600, self.getLib().dateValue(self.makeDate("-1970-01-02")))
        # self.assertEqual(-124331673600, self.getLib().dateValue(self.makeDate("-1970-02-01")))
        # self.assertEqual(-125281123200, self.getLib().dateValue(self.makeDate("-2000-01-01")))

    def testDateIs(self):
        self.assertTrue(self.getLib().dateIs(None, None))
        self.assertFalse(self.getLib().dateIs(None, self.makeDate("2016-08-01")))
        self.assertFalse(self.getLib().dateIs(self.makeDate("2016-08-01"), None))

        self.assertTrue(self.getLib().dateIs(self.makeDate("2016-08-01"), self.makeDate("2016-08-01")))
        self.assertFalse(self.getLib().dateIs(self.makeDate("2016-08-01"), self.makeDate("2016-08-02")))

    def testDateEqual(self):
        self.assertTrue(self.getLib().dateEqual(None, None))
        self.assertFalse(self.getLib().dateEqual(None, self.makeDate("2016-08-01")))
        self.assertFalse(self.getLib().dateEqual(self.makeDate("2016-08-01"), None))

        self.assertTrue(self.getLib().dateEqual(self.makeDate("2016-08-01"), self.makeDate("2016-08-01")))
        self.assertFalse(self.getLib().dateEqual(self.makeDate("2016-08-01"), self.makeDate("2016-08-02")))

    def testDateNotEqual(self):
        self.assertFalse(self.getLib().dateNotEqual(None, None))
        self.assertTrue(self.getLib().dateNotEqual(None, self.makeDate("2016-08-01")))
        self.assertTrue(self.getLib().dateNotEqual(self.makeDate("2016-08-01"), None))

        self.assertFalse(self.getLib().dateNotEqual(self.makeDate("2016-08-01"), self.makeDate("2016-08-01")))
        self.assertTrue(self.getLib().dateEqual(self.makeDate("2016-08-01"), self.makeDate("2016-08-01")))

    def testDateLessThan(self):
        self.assertIsNone(self.getLib().dateLessThan(None, None))
        self.assertIsNone(self.getLib().dateLessThan(None, self.makeDate("2016-08-01")))
        self.assertIsNone(self.getLib().dateLessThan(self.makeDate("2016-08-01"), None))

        self.assertFalse(self.getLib().dateLessThan(self.makeDate("2016-08-01"), self.makeDate("2016-08-01")))
        self.assertTrue(self.getLib().dateLessThan(self.makeDate("2016-08-01"), self.makeDate("2016-08-02")))

    def testDateGreaterThan(self):
        self.assertIsNone(self.getLib().dateGreaterThan(None, None))
        self.assertIsNone(self.getLib().dateGreaterThan(None, self.makeDate("2016-08-01")))
        self.assertIsNone(self.getLib().dateGreaterThan(self.makeDate("2016-08-01"), None))

        self.assertFalse(self.getLib().dateGreaterThan(self.makeDate("2016-08-01"), self.makeDate("2016-08-01")))
        self.assertTrue(self.getLib().dateGreaterThan(self.makeDate("2016-08-02"), self.makeDate("2016-08-01")))

    def testDateLessEqualThan(self):
        self.assertTrue(self.getLib().dateLessEqualThan(None, None))
        self.assertIsNone(self.getLib().dateLessEqualThan(None, self.makeDate("2016-08-01")))
        self.assertIsNone(self.getLib().dateLessEqualThan(self.makeDate("2016-08-01"), None))

        self.assertTrue(self.getLib().dateLessEqualThan(self.makeDate("2016-08-01"), self.makeDate("2016-08-01")))
        self.assertFalse(self.getLib().dateLessEqualThan(self.makeDate("2016-08-03"), self.makeDate("2016-08-02")))

    def testDateGreaterEqualThan(self):
        self.assertTrue(self.getLib().dateGreaterEqualThan(None, None))
        self.assertIsNone(self.getLib().dateGreaterEqualThan(None, self.makeDate("2016-08-01")))
        self.assertIsNone(self.getLib().dateGreaterEqualThan(self.makeDate("2016-08-01"), None))

        self.assertTrue(self.getLib().dateGreaterEqualThan(self.makeDate("2016-08-01"), self.makeDate("2016-08-01")))
        self.assertFalse(self.getLib().dateGreaterEqualThan(self.makeDate("2016-08-01"), self.makeDate("2016-08-03")))

    def testDateSubtract(self):
        self.assertIsNone(self.getLib().dateSubtract(None, None))
        self.assertIsNone(self.getLib().dateSubtract(None, self.makeDate("2016-08-01")))
        self.assertIsNone(self.getLib().dateSubtract(self.makeDate("2016-08-01"), None))

        self.assertEqual(self.makeDuration("PT0S"), self.getLib().dateSubtract(self.makeDate("2016-08-01"), self.makeDate("2016-08-01")))
        self.assertEqual(self.makeDuration("-P2D"), self.getLib().dateSubtract(self.makeDate("2016-08-01"), self.makeDate("2016-08-03")))

    def testDateAddDuration(self):
        self.assertIsNone(self.getLib().dateAddDuration(None, None))
        self.assertIsNone(self.getLib().dateAddDuration(None, self.makeDuration("P0Y2M")))
        self.assertIsNone(self.getLib().dateAddDuration(self.makeDate("2016-08-01"), None))

        self.assertEqualsDate("2016-10-01", self.getLib().dateAddDuration(self.makeDate("2016-08-01"), self.makeDuration("P0Y2M")))
        self.assertEqualsDate("2016-06-01", self.getLib().dateAddDuration(self.makeDate("2016-08-01"), self.makeDuration("-P0Y2M")))

    def testDateSubtractDuration(self):
        self.assertIsNone(self.getLib().dateSubtractDuration(None, None))
        self.assertIsNone(self.getLib().dateSubtractDuration(None, self.makeDuration("P0Y2M")))
        self.assertIsNone(self.getLib().dateSubtractDuration(self.makeDate("2016-08-01"), None))

        self.assertEqualsDate("2016-06-01", self.getLib().dateSubtractDuration(self.makeDate("2016-08-01"), self.makeDuration("P0Y2M")))
        self.assertEqualsDate("2016-10-01", self.getLib().dateSubtractDuration(self.makeDate("2016-08-01"), self.makeDuration("-P0Y2M")))

    #
    # Time operators
    #
    def testIsTime(self):
        self.assertFalse(self.getLib().isTime(None))
        self.assertFalse(self.getLib().isTime(self.getLib().number("1")))
        self.assertFalse(self.getLib().isTime("abc"))
        self.assertFalse(self.getLib().isTime(True))
        self.assertFalse(self.getLib().isTime(self.getLib().date("2020-01-01")))
        self.assertTrue(self.getLib().isTime(self.getLib().time("12:00:00")))
        self.assertFalse(self.getLib().isTime(self.getLib().dateAndTime("2020-01-01T12:00:00")))
        self.assertFalse(self.getLib().isTime(self.getLib().duration("P1Y1M")))
        self.assertFalse(self.getLib().isTime(self.getLib().asList("a")))
        self.assertFalse(self.getLib().isTime(Context()))
        self.assertFalse(self.getLib().isTime(Range(True, 0, True, 1)))

    def testTimeValue(self):
        self.assertIsNone(self.getLib().timeValue(None))

        # local time
        self.assertEqual(3723, self.getLib().timeValue(self.makeTime("01:02:03")))
        self.assertEqual(3723, self.getLib().timeValue(self.makeTime("01:02:03.0004")))

        # offset time
        self.assertEqual(3723, self.getLib().timeValue(self.makeTime("01:02:03Z")))
        self.assertEqual(3723, self.getLib().timeValue(self.makeTime("01:02:03Z")))
        self.assertEqual(3723, self.getLib().timeValue(self.makeTime("01:02:03+00:00")))
        self.assertEqual(63, self.getLib().timeValue(self.makeTime("01:02:03+01:01")))

        # zoneid time
        self.assertEqual(123, self.getLib().timeValue(self.makeTime("01:02:03@Europe/Paris")))
        self.assertEqual(3723, self.getLib().timeValue(self.makeTime("01:02:03@Etc/UTC")))

    def testTimeIs(self):
        self.assertTrue(self.getLib().timeIs(None, None))
        self.assertFalse(self.getLib().timeIs(None, self.makeTime("12:00:00Z")))
        self.assertFalse(self.getLib().timeIs(self.makeTime("12:00:00Z"), None))

        # same times are is()
        self.assertTrue(self.getLib().timeIs(self.makeTime("10:30:00"), self.makeTime("10:30:00")))
        # different times are not is()
        self.assertFalse(self.getLib().timeIs(self.makeTime("10:30:00"), self.makeTime("10:30:01")))
        # different times with zero milliseconds are is()
        self.assertTrue(self.getLib().timeIs(self.makeTime("10:30:00.0000"), self.makeTime("10:30:00")))
        # different times with same milliseconds are is()
        self.assertTrue(self.getLib().timeIs(self.makeTime("10:30:00.0001"), self.makeTime("10:30:00.0001")))
        # different times with different milliseconds are is()
        self.assertTrue(self.getLib().timeIs(self.makeTime("10:30:00.0001"), self.makeTime("10:30:00.0002")))
        # same times in same zone are is()
        self.assertTrue(self.getLib().timeIs(self.makeTime("10:30:00@Europe/Paris"), self.makeTime("10:30:00@Europe/Paris")))
        # same times - one with zone one without are not is()
        self.assertFalse(self.getLib().timeIs(self.makeTime("10:30:00@Europe/Paris"), self.makeTime("10:30:00")))
        # same times with different zones are not is()
        self.assertFalse(self.getLib().timeIs(self.makeTime("10:30:00@Europe/Paris"), self.makeTime("10:30:00@Asia/Dhaka")))
        # same times = one with offset, the other with zone are not equal
        self.assertFalse(self.getLib().timeIs(self.makeTime("10:30:00+02:00"), self.makeTime("10:30:00@Europe/Paris")))
        # same times = one with Z zone, the other with UTC are is()
        self.assertTrue(self.getLib().timeIs(self.makeTime("10:30:00Z"), self.makeTime("10:30:00+00:00")))

    def testTimeEqual(self):
        self.assertTrue(self.getLib().timeEqual(None, None))
        self.assertFalse(self.getLib().timeEqual(None, self.makeTime("12:00:00Z")))
        self.assertFalse(self.getLib().timeEqual(self.makeTime("12:00:00Z"), None))

        # same times are equal
        self.assertTrue(self.getLib().timeEqual(self.makeTime("10:30:00"), self.makeTime("10:30:00")))
        # different times are not equal
        self.assertFalse(self.getLib().timeEqual(self.makeTime("10:30:00"), self.makeTime("10:30:01")))
        # same times with zero milliseconds are equal
        self.assertTrue(self.getLib().timeEqual(self.makeTime("10:30:00.0000"), self.makeTime("10:30:00")))
        # same times with same milliseconds are equal
        self.assertTrue(self.getLib().timeEqual(self.makeTime("10:30:00.0001"), self.makeTime("10:30:00.0001")))
        # same times with different milliseconds are equal
        self.assertTrue(self.getLib().timeEqual(self.makeTime("10:30:00.0001"), self.makeTime("10:30:00.0002")))
        # same times in same zone are equal
        self.assertTrue(self.getLib().timeEqual(self.makeTime("10:30:00@Europe/Paris"), self.makeTime("10:30:00@Europe/Paris")))
        # same times - one with zone one without are not equal
        self.assertFalse(self.getLib().timeEqual(self.makeTime("10:30:00@Europe/Paris"), self.makeTime("10:30:00")))
        # same times with different zones are not equal
        self.assertFalse(self.getLib().timeEqual(self.makeTime("10:30:00@Europe/Paris"), self.makeTime("10:30:00@Asia/Dhaka")))
        # same times = one with offset, the other with zone are not equal
        self.assertFalse(self.getLib().timeEqual(self.makeTime("10:30:00+02:00"), self.makeTime("10:30:00@Europe/Paris")))
        # same times = one with Z zone, the other with UTC are equal
        self.assertTrue(self.getLib().timeEqual(self.makeTime("10:30:00Z"), self.makeTime("10:30:00+00:00")))

        # times with equivalent offset and zone id are equal
        self.assertTrue(self.getLib().timeEqual(self.makeTime("12:00:00"), self.makeTime("12:00:00+00:00")))
        self.assertTrue(self.getLib().timeEqual(self.makeTime("00:00:00+00:00"), self.makeTime("00:00:00@Etc/UTC")))
        self.assertTrue(self.getLib().timeEqual(self.makeTime("00:00:00Z"), self.makeTime("00:00:00+00:00")))
        self.assertTrue(self.getLib().timeEqual(self.makeTime("00:00:00Z"), self.makeTime("00:00:00@Etc/UTC")))

    def testTimeNotEqual(self):
        self.assertFalse(self.getLib().timeNotEqual(None, None))
        self.assertTrue(self.getLib().timeNotEqual(None, self.makeTime("12:00:00Z")))
        self.assertTrue(self.getLib().timeNotEqual(self.makeTime("12:00:00Z"), None))

        self.assertFalse(self.getLib().timeNotEqual(self.makeTime("12:00:00Z"), self.makeTime("12:00:00Z")))
        self.assertTrue(self.getLib().timeNotEqual(self.makeTime("12:00:00Z"), self.makeTime("12:00:01Z")))

    def testTimeLessThan(self):
        self.assertIsNone(self.getLib().timeLessThan(None, None))
        self.assertIsNone(self.getLib().timeLessThan(None, self.makeTime("12:00:00Z")))
        self.assertIsNone(self.getLib().timeLessThan(self.makeTime("12:00:00Z"), None))

        self.assertFalse(self.getLib().timeLessThan(self.makeTime("12:00:00Z"), self.makeTime("12:00:00Z")))
        self.assertTrue(self.getLib().timeLessThan(self.makeTime("11:00:00Z"), self.makeTime("12:00:01Z")))

    def testTimeGreaterThan(self):
        self.assertIsNone(self.getLib().timeGreaterThan(None, None))
        self.assertIsNone(self.getLib().timeGreaterThan(None, self.makeTime("12:00:00Z")))
        self.assertIsNone(self.getLib().timeGreaterThan(self.makeTime("12:00:00Z"), None))

        self.assertFalse(self.getLib().timeGreaterThan(self.makeTime("12:00:00Z"), self.makeTime("12:00:00Z")))
        self.assertTrue(self.getLib().timeGreaterThan(self.makeTime("13:00:00Z"), self.makeTime("12:00:01Z")))

    def testTimeLessEqualThan(self):
        self.assertTrue(self.getLib().timeLessEqualThan(None, None))
        self.assertIsNone(self.getLib().timeLessEqualThan(None, self.makeTime("12:00:00Z")))
        self.assertIsNone(self.getLib().timeLessEqualThan(self.makeTime("12:00:00Z"), None))

        self.assertTrue(self.getLib().timeLessEqualThan(self.makeTime("12:00:00Z"), self.makeTime("12:00:00Z")))
        self.assertFalse(self.getLib().timeLessEqualThan(self.makeTime("13:00:00Z"), self.makeTime("12:00:01Z")))

    def testTimeGreaterEqualThan(self):
        self.assertTrue(self.getLib().timeGreaterEqualThan(None, None))
        self.assertIsNone(self.getLib().timeGreaterEqualThan(None, self.makeTime("12:00:00Z")))
        self.assertIsNone(self.getLib().timeGreaterEqualThan(self.makeTime("12:00:00Z"), None))

        self.assertTrue(self.getLib().timeGreaterEqualThan(self.makeTime("12:00:00Z"), self.makeTime("12:00:00Z")))
        self.assertFalse(self.getLib().timeGreaterEqualThan(self.makeTime("11:00:00Z"), self.makeTime("12:00:01Z")))

    def testTimeSubtract(self):
        self.assertIsNone(self.getLib().timeSubtract(None, None))
        self.assertIsNone(self.getLib().timeSubtract(None, self.makeTime("12:00:00Z")))
        self.assertIsNone(self.getLib().timeSubtract(self.makeTime("12:00:00Z"), None))

        self.assertEqual(self.makeDuration("PT1H"), self.getLib().timeSubtract(self.makeTime("13:00:00Z"), self.makeTime("12:00:00Z")))
        self.assertEqual(self.makeDuration("P0D"), self.getLib().timeSubtract(self.makeTime("12:00:00Z"), self.makeTime("12:00:00Z")))
        self.assertEqual(self.makeDuration("-PT1H"), self.getLib().timeSubtract(self.makeTime("12:00:00Z"), self.makeTime("13:00:00Z")))

    def testTimeAddDuration(self):
        self.assertIsNone(self.getLib().timeAddDuration(None, None))
        self.assertIsNone(self.getLib().timeAddDuration(None, self.makeDuration("P0DT1H")))
        self.assertIsNone(self.getLib().timeAddDuration(self.makeTime("12:00:00Z"), None))

        self.assertEqualsTime("13:00:01Z", self.getLib().timeAddDuration(self.makeTime("12:00:01Z"), self.makeDuration("P0DT1H")))
        self.assertEqualsTime("12:00:01Z", self.getLib().timeAddDuration(self.makeTime("12:00:01Z"), self.makeDuration("P1DT0H")))

    def testTimeSubtractDuration(self):
        self.assertIsNone(self.getLib().timeSubtractDuration(None, None))
        self.assertIsNone(self.getLib().timeSubtractDuration(None, self.makeDuration("P0DT1H")))
        self.assertIsNone(self.getLib().timeSubtractDuration(self.makeTime("12:00:01Z"), None))

        self.assertEqualsTime("11:00:01Z", self.getLib().timeSubtractDuration(self.makeTime("12:00:01Z"), self.makeDuration("P0DT1H")))
        self.assertEqualsTime("12:00:01Z", self.getLib().timeSubtractDuration(self.makeTime("12:00:01Z"), self.makeDuration("P1DT0H")))

    #
    # Date and time operators
    #
    def testIsDateTime(self):
        self.assertFalse(self.getLib().isDateTime(None))
        self.assertFalse(self.getLib().isDateTime(self.getLib().number("1")))
        self.assertFalse(self.getLib().isDateTime("abc"))
        self.assertFalse(self.getLib().isDateTime(True))
        self.assertFalse(self.getLib().isDateTime(self.getLib().date("2020-01-01")))
        self.assertFalse(self.getLib().isDateTime(self.getLib().time("12:00:00")))
        self.assertTrue(self.getLib().isDateTime(self.getLib().dateAndTime("2020-01-01T12:00:00")))
        self.assertFalse(self.getLib().isDateTime(self.getLib().duration("P1Y1M")))
        self.assertFalse(self.getLib().isDateTime(self.getLib().asList("a")))
        self.assertFalse(self.getLib().isDateTime(Context()))
        self.assertFalse(self.getLib().isDateTime(Range(True, 0, True, 1)))

    def testDateTimeValue(self):
        self.assertIsNone(self.getLib().dateTimeValue(None))

        # local date time
        self.assertEqual(3723, self.getLib().dateTimeValue(self.makeDateAndTime("1970-01-01T01:02:03")))
        self.assertEqual(3723, self.getLib().dateTimeValue(self.makeDateAndTime("1970-01-01T01:02:03.0004")))
        self.assertEqual(315536523, self.getLib().dateTimeValue(self.makeDateAndTime("1980-01-01T01:02:03.0004")))
#        self.assertEqual(-124649967477, self.getLib().dateTimeValue(self.makeDateAndTime("-1980-01-01T01:02:03.0004")))

        # offset date time
        self.assertEqual(3723, self.getLib().dateTimeValue(self.makeDateAndTime("1970-01-01T01:02:03Z")))
        self.assertEqual(3723, self.getLib().dateTimeValue(self.makeDateAndTime("1970-01-01T01:02:03Z")))
        self.assertEqual(3723, self.getLib().dateTimeValue(self.makeDateAndTime("1970-01-01T01:02:03+00:00")))
        self.assertEqual(63, self.getLib().dateTimeValue(self.makeDateAndTime("1970-01-01T01:02:03+01:01")))
        self.assertEqual(315532863, self.getLib().dateTimeValue(self.makeDateAndTime("1980-01-01T01:02:03+01:01")))
#        self.assertEqual(-124649971137, self.getLib().dateTimeValue(self.makeDateAndTime("-1980-01-01T01:02:03+01:01")))

        # zoneid date time
        self.assertEqual(123, self.getLib().dateTimeValue(self.makeDateAndTime("1970-01-01T01:02:03@Europe/Paris")))
        self.assertEqual(3723, self.getLib().dateTimeValue(self.makeDateAndTime("1970-01-01T01:02:03@Etc/UTC")))
        self.assertEqual(315536523, self.getLib().dateTimeValue(self.makeDateAndTime("1980-01-01T01:02:03@Etc/UTC")))
#        self.assertEqual(-124649967477, self.getLib().dateTimeValue(self.makeDateAndTime("-1980-01-01T01:02:03@Etc/UTC")))

    def testDateTimeIs(self):
        # datetime equals None
        self.assertTrue(self.getLib().dateTimeIs(None, None))
        self.assertFalse(self.getLib().dateTimeIs(self.makeDateAndTime("2018-12-08T00:00:00"), None))
        self.assertFalse(self.getLib().dateTimeIs(None, self.makeDateAndTime("2018-12-08T00:00:00")))

        self.assertTrue(self.getLib().dateTimeIs(self.makeDateAndTime("2016-08-01T11:00:00Z"), self.makeDateAndTime("2016-08-01T11:00:00Z")))
        self.assertFalse(self.getLib().dateTimeIs(self.makeDateAndTime("2016-08-01T11:00:00Z"), self.makeDateAndTime("2016-08-01T11:00:01Z")))

        # same datetimes are is ()
        self.assertTrue(self.getLib().dateTimeIs(self.makeDateAndTime("2018-12-08T10:30:00"), self.makeDateAndTime("2018-12-08T10:30:00")))
        # datetimes with no time is is () to datetime with zero time
        self.assertTrue(self.getLib().dateTimeIs(self.makeDateAndTime("2018-12-08"), self.makeDateAndTime("2018-12-08T00:00:00")))
        # datetimes with same milliseconds are is ()
        self.assertTrue(self.getLib().dateTimeIs(self.makeDateAndTime("2018-12-08T00:00:00.0001"), self.makeDateAndTime("2018-12-08T00:00:00.0001")))
        # datetimes with different milliseconds are is ()
        self.assertTrue(self.getLib().dateTimeIs(self.makeDateAndTime("2018-12-08T00:00:00.0001"), self.makeDateAndTime("2018-12-08T00:00:00.0002")))
        # different datetimes are not is ()
        self.assertFalse(self.getLib().dateTimeIs(self.makeDateAndTime("2018-12-08T00:00:00"), self.makeDateAndTime("2018-12-07T00:00:00")))
        # same datetimes in same zone are is ()
        self.assertTrue(self.getLib().dateTimeIs(self.makeDateAndTime("2018-12-08T00:00:00@Europe/Paris"), self.makeDateAndTime("2018-12-08T00:00:00@Europe/Paris")))
        # same datetimes in different zones are not is ()
        self.assertFalse(self.getLib().dateTimeIs(self.makeDateAndTime("2018-12-08T00:00:00@Europe/Paris"), self.makeDateAndTime("2018-12-08T00:00:00@Asia/Dhaka")))
        # same datetimes, one with zone one without are not is ()
        self.assertFalse(self.getLib().dateTimeIs(self.makeDateAndTime("2018-12-08T00:00:00"), self.makeDateAndTime("2018-12-08T00:00:00@Asia/Dhaka")))
        # same datetimes, one with offset and the other with zone are not is ()
        self.assertFalse(self.getLib().dateTimeIs(self.makeDateAndTime("2018-12-08T00:00:00+02:00"), self.makeDateAndTime("2018-12-08T00:00:00@Europe/Paris")))

    def testDateTimeEqual(self):
        # datetime equals None
        self.assertTrue(self.getLib().dateTimeEqual(None, None))
        self.assertFalse(self.getLib().dateTimeEqual(self.makeDateAndTime("2018-12-08T00:00:00"), None))
        self.assertFalse(self.getLib().dateTimeEqual(None, self.makeDateAndTime("2018-12-08T00:00:00")))

        self.assertTrue(self.getLib().dateTimeEqual(self.makeDateAndTime("2016-08-01T11:00:00Z"), self.makeDateAndTime("2016-08-01T11:00:00Z")))
        self.assertFalse(self.getLib().dateTimeEqual(self.makeDateAndTime("2016-08-01T11:00:00Z"), self.makeDateAndTime("2016-08-01T11:00:01Z")))

        # same datetimes are equal
        self.assertTrue(self.getLib().dateTimeEqual(self.makeDateAndTime("2018-12-08T10:30:00"), self.makeDateAndTime("2018-12-08T10:30:00")))
        # datetimes with no time is equal to datetime with zero time
        self.assertTrue(self.getLib().dateTimeEqual(self.makeDateAndTime("2018-12-08"), self.makeDateAndTime("2018-12-08T00:00:00")))
        # datetimes with same milliseconds are equal
        self.assertTrue(self.getLib().dateTimeEqual(self.makeDateAndTime("2018-12-08T00:00:00.0001"), self.makeDateAndTime("2018-12-08T00:00:00.0001")))
        # datetimes with different milliseconds are equal
        self.assertTrue(self.getLib().dateTimeEqual(self.makeDateAndTime("2018-12-08T00:00:00.0001"), self.makeDateAndTime("2018-12-08T00:00:00.0002")))
        # different datetimes are not equal
        self.assertFalse(self.getLib().dateTimeEqual(self.makeDateAndTime("2018-12-08T00:00:00"), self.makeDateAndTime("2018-12-07T00:00:00")))
        # same datetimes in same zone are equal
        self.assertTrue(self.getLib().dateTimeEqual(self.makeDateAndTime("2018-12-08T00:00:00@Europe/Paris"), self.makeDateAndTime("2018-12-08T00:00:00@Europe/Paris")))
        # same datetimes in different zones are not equal
        self.assertFalse(self.getLib().dateTimeEqual(self.makeDateAndTime("2018-12-08T00:00:00@Europe/Paris"), self.makeDateAndTime("2018-12-08T00:00:00@Asia/Dhaka")))
        # same datetimes, one with zone one without are not equal
        self.assertFalse(self.getLib().dateTimeEqual(self.makeDateAndTime("2018-12-08T00:00:00"), self.makeDateAndTime("2018-12-08T00:00:00@Asia/Dhaka")))
        # same datetimes, one with offset and the other with zone are not equal
        self.assertFalse(self.getLib().dateTimeEqual(self.makeDateAndTime("2018-12-08T00:00:00+02:00"), self.makeDateAndTime("2018-12-08T00:00:00@Europe/Paris")))

        # datetime with equivalent offset and zone id are equal
        self.assertTrue(self.getLib().dateTimeEqual(self.makeDateAndTime("2018-12-08T00:00:00+00:00"), self.makeDateAndTime("2018-12-08T00:00:00@Etc/UTC")))
        self.assertTrue(self.getLib().dateTimeEqual(self.makeDateAndTime("2018-12-08T12:00:00Z"), self.makeDateAndTime("2018-12-08T12:00:00+00:00")))
        self.assertTrue(self.getLib().dateTimeEqual(self.makeDateAndTime("2018-12-08T00:00:00Z"), self.makeDateAndTime("2018-12-08T00:00:00@Etc/UTC")))

    def testDateTimeNotEqual(self):
        self.assertFalse(self.getLib().dateTimeNotEqual(None, None))
        self.assertTrue(self.getLib().dateTimeNotEqual(None, self.makeDateAndTime("2016-08-01T11:00:00Z")))
        self.assertTrue(self.getLib().dateTimeNotEqual(self.makeDateAndTime("2016-08-01T11:00:00Z"), None))

        self.assertFalse(self.getLib().dateTimeNotEqual(self.makeDateAndTime("2016-08-01T11:00:00Z"), self.makeDateAndTime("2016-08-01T11:00:00Z")))
        self.assertTrue(self.getLib().dateTimeNotEqual(self.makeDateAndTime("2016-08-01T11:00:00Z"), self.makeDateAndTime("2016-08-01T11:00:01Z")))

    def testDateTimeLessThan(self):
        self.assertIsNone(self.getLib().dateTimeLessThan(None, None))
        self.assertIsNone(self.getLib().dateTimeLessThan(None, self.makeDateAndTime("2016-08-01T11:00:00Z")))
        self.assertIsNone(self.getLib().dateTimeLessThan(self.makeDateAndTime("2016-08-01T11:00:00Z"), None))

        self.assertFalse(self.getLib().dateTimeLessThan(self.makeDateAndTime("2016-08-01T11:00:00Z"), self.makeDateAndTime("2016-08-01T11:00:00Z")))
        self.assertTrue(self.getLib().dateTimeLessThan(self.makeDateAndTime("2016-08-01T11:00:00Z"), self.makeDateAndTime("2017-08-01T11:00:01Z")))

    def testDateTimeGreaterThan(self):
        self.assertIsNone(self.getLib().dateTimeGreaterThan(None, None))
        self.assertIsNone(self.getLib().dateTimeGreaterThan(None, self.makeDateAndTime("2016-08-01T11:00:00Z")))
        self.assertIsNone(self.getLib().dateTimeGreaterThan(self.makeDateAndTime("2016-08-01T11:00:00Z"), None))

        self.assertTrue(self.getLib().dateTimeGreaterThan(self.makeDateAndTime("2017-08-01T11:00:00Z"), self.makeDateAndTime("2016-08-01T11:00:00Z")))
        self.assertFalse(self.getLib().dateTimeGreaterThan(self.makeDateAndTime("2016-08-01T11:00:00Z"), self.makeDateAndTime("2016-08-01T11:00:01Z")))

    def testDateTimeLessEqualThan(self):
        self.assertTrue(self.getLib().dateTimeLessEqualThan(None, None))
        self.assertIsNone(self.getLib().dateTimeLessEqualThan(None, self.makeDateAndTime("2016-08-01T11:00:00Z")))
        self.assertIsNone(self.getLib().dateTimeLessEqualThan(self.makeDateAndTime("2016-08-01T11:00:00Z"), None))

        self.assertTrue(self.getLib().dateTimeLessEqualThan(self.makeDateAndTime("2016-08-01T11:00:00Z"), self.makeDateAndTime("2016-08-01T11:00:00Z")))
        self.assertFalse(self.getLib().dateTimeLessEqualThan(self.makeDateAndTime("2016-08-01T11:00:01Z"), self.makeDateAndTime("2016-08-01T11:00:00Z")))

    def testDateTimeGreaterEqualThan(self):
        self.assertTrue(self.getLib().dateTimeGreaterEqualThan(None, None))
        self.assertIsNone(self.getLib().dateTimeGreaterEqualThan(None, self.makeDateAndTime("2016-08-01T11:00:00Z")))
        self.assertIsNone(self.getLib().dateTimeGreaterEqualThan(self.makeDateAndTime("2016-08-01T11:00:00Z"), None))

        self.assertTrue(self.getLib().dateTimeGreaterEqualThan(self.makeDateAndTime("2016-08-01T11:00:00Z"), self.makeDateAndTime("2016-08-01T11:00:00Z")))
        self.assertFalse(self.getLib().dateTimeGreaterEqualThan(self.makeDateAndTime("2015-08-01T11:00:00Z"), self.makeDateAndTime("2016-08-01T11:00:01Z")))

    def testDateTimeSubtract(self):
        self.assertIsNone(self.getLib().dateTimeSubtract(None, None))
        self.assertIsNone(self.getLib().dateTimeSubtract(None, self.makeDateAndTime("2016-08-01T12:00:00Z")))
        self.assertIsNone(self.getLib().dateTimeSubtract(self.makeDateAndTime("2016-08-01T12:00:00Z"), None))

        self.assertEqual(self.makeDuration("PT1H"), self.getLib().dateTimeSubtract(self.makeDateAndTime("2016-08-01T13:00:00Z"), self.makeDateAndTime("2016-08-01T12:00:00Z")))
        self.assertEqual(self.makeDuration("PT0S"), self.getLib().dateTimeSubtract(self.makeDateAndTime("2016-08-01T12:00:00Z"), self.makeDateAndTime("2016-08-01T12:00:00Z")))
        self.assertEqual(self.makeDuration("-P2DT1H"), self.getLib().dateTimeSubtract(self.makeDateAndTime("2016-08-01T12:00:00Z"), self.makeDateAndTime("2016-08-03T13:00:00Z")))

        self.assertEqual(self.makeDuration("P367DT6H58M59S"), self.getLib().dateTimeSubtract(
            self.makeDateAndTime("2016-12-24T23:59:00-08:00"), self.makeDateAndTime("2015-12-24T00:00:01-01:00")))

    def testDateTimeAddDuration(self):
        self.assertIsNone(self.getLib().dateTimeAddDuration(None, None))
        self.assertIsNone(self.getLib().dateTimeAddDuration(None, self.makeDuration("P1Y1M")))
        self.assertIsNone(self.getLib().dateTimeAddDuration(self.makeDateAndTime("2016-08-01T12:00:00Z"), None))

        self.assertEqualsDateTime("2017-03-01T12:00:01Z", self.getLib().dateTimeAddDuration(self.makeDateAndTime("2016-02-01T12:00:01Z"), self.makeDuration("P1Y1M")))
        self.assertEqualsDateTime("2015-01-01T12:00:01Z", self.getLib().dateTimeAddDuration(self.makeDateAndTime("2016-02-01T12:00:01Z"), self.makeDuration("-P1Y1M")))

        self.assertEqualsDateTime("2016-02-02T13:00:01Z", self.getLib().dateTimeAddDuration(self.makeDateAndTime("2016-02-01T12:00:01Z"), self.makeDuration("P1DT1H")))
        self.assertEqualsDateTime("2016-01-31T11:00:01Z", self.getLib().dateTimeAddDuration(self.makeDateAndTime("2016-02-01T12:00:01Z"), self.makeDuration("-P1DT1H")))

    def testDateTimeSubtractDuration(self):
        self.assertIsNone(self.getLib().dateTimeSubtractDuration(None, None))
        self.assertIsNone(self.getLib().dateTimeSubtractDuration(None, self.makeDuration("P1Y1M")))
        self.assertIsNone(self.getLib().dateTimeSubtractDuration(self.makeDateAndTime("2016-08-01T12:00:00Z"), None))

        self.assertEqualsDateTime("2015-01-01T12:00:01Z", self.getLib().dateTimeSubtractDuration(self.makeDateAndTime("2016-02-01T12:00:01Z"), self.makeDuration("P1Y1M")))
        self.assertEqualsDateTime("2017-03-01T12:00:01Z", self.getLib().dateTimeSubtractDuration(self.makeDateAndTime("2016-02-01T12:00:01Z"), self.makeDuration("-P1Y1M")))

        self.assertEqualsDateTime("2016-01-31T11:00:01Z", self.getLib().dateTimeSubtractDuration(self.makeDateAndTime("2016-02-01T12:00:01Z"), self.makeDuration("P1DT1H")))
        self.assertEqualsDateTime("2016-02-02T13:00:01Z", self.getLib().dateTimeSubtractDuration(self.makeDateAndTime("2016-02-01T12:00:01Z"), self.makeDuration("-P1DT1H")))

    #
    # Duration operators
    #
    def testIsDuration(self):
        self.assertFalse(self.getLib().isDuration(None))

        # years and months
        self.assertTrue(self.getLib().isYearsAndMonthsDuration(self.makeDuration("P1Y2M")))
        self.assertTrue(self.getLib().isYearsAndMonthsDuration(self.makeDuration("-P1Y2M")))
        self.assertFalse(self.getLib().isYearsAndMonthsDuration(self.makeDuration("P1DT2H3M4S")))
        self.assertFalse(self.getLib().isYearsAndMonthsDuration(self.makeDuration("-P1DT2H3M4S")))
        self.assertFalse(self.getLib().isYearsAndMonthsDuration(self.makeDuration("P1Y2M1DT2H3M4S")))
        self.assertFalse(self.getLib().isYearsAndMonthsDuration(self.makeDuration("-P1Y2M1DT2H3M4S")))

        # days and time
        self.assertFalse(self.getLib().isDaysAndTimeDuration(self.makeDuration("P1Y2M")))
        self.assertFalse(self.getLib().isDaysAndTimeDuration(self.makeDuration("-P1Y2M")))
        self.assertTrue(self.getLib().isDaysAndTimeDuration(self.makeDuration("P1DT2H3M4S")))
        self.assertTrue(self.getLib().isDaysAndTimeDuration(self.makeDuration("-P1DT2H3M4S")))
        self.assertFalse(self.getLib().isDaysAndTimeDuration(self.makeDuration("P1Y2M1DT2H3M4S")))
        self.assertFalse(self.getLib().isDaysAndTimeDuration(self.makeDuration("-P1Y2M1DT2H3M4S")))

        # mixture
        self.assertTrue(self.getLib().isDuration(self.makeDuration("P1Y2M")))
        self.assertTrue(self.getLib().isDuration(self.makeDuration("-P1Y2M")))
        self.assertTrue(self.getLib().isDuration(self.makeDuration("P1DT2H3M4S")))
        self.assertTrue(self.getLib().isDuration(self.makeDuration("-P1DT2H3M4S")))
        self.assertTrue(self.getLib().isDuration(self.makeDuration("P1Y2M1DT2H3M4S")))
        self.assertTrue(self.getLib().isDuration(self.makeDuration("-P1Y2M1DT2H3M4S")))

    def testDurationValue(self):
        self.assertIsNone(self.getLib().durationValue(None))

        # years and months
        self.assertEqual(12 + 2, self.getLib().durationValue(self.makeDuration("P1Y2M")))
        self.assertEqual(-(12 + 2), self.getLib().durationValue(self.makeDuration("-P1Y2M")))

        # days and time
        self.assertEqual((24 + 2) * 3600 + 3 * 60 + 4, self.getLib().durationValue(self.makeDuration("P1DT2H3M4S")))
        self.assertEqual(- ((24 + 2) * 3600 + 3 * 60 + 4), self.getLib().durationValue(self.makeDuration("-P1DT2H3M4S")))

        # mixture
        self.assertEqual(36727384, self.getLib().durationValue(self.makeDuration("P1Y2M1DT2H3M4S")))
#        self.assertEqual(-36727384, self.getLib().durationValue(self.makeDuration("-P1Y2M1DT2H3M4S")))

    def testDurationIs(self):
        self.assertTrue(self.getLib().durationIs(None, None))
        self.assertFalse(self.getLib().durationIs(None, self.makeDuration("P1Y1M")))
        self.assertFalse(self.getLib().durationIs(self.makeDuration("P1Y1M"), None))

        self.assertTrue(self.getLib().durationIs(self.makeDuration("P1Y1M"), self.makeDuration("P1Y1M")))
        self.assertTrue(self.getLib().durationIs(self.makeDuration("P1Y1M"), self.makeDuration("P0Y13M")))
        self.assertFalse(self.getLib().durationIs(self.makeDuration("P1Y1M"), self.makeDuration("P1Y2M")))

        self.assertTrue(self.getLib().durationIs(self.makeDuration("P1DT1H"), self.makeDuration("P1DT1H")))
        self.assertTrue(self.getLib().durationIs(self.makeDuration("P1DT1H"), self.makeDuration("P0DT25H")))
        self.assertFalse(self.getLib().durationIs(self.makeDuration("P1DT1H"), self.makeDuration("P1DT2H")))

        # different semantic domains
        # both are timedelta
#        self.assertFalse(self.getLib().durationIs(self.makeDuration("P0Y"), self.makeDuration("P0D")))

    def testDurationEqual(self):
        self.assertTrue(self.getLib().durationEqual(None, None))
        self.assertFalse(self.getLib().durationEqual(None, self.makeDuration("P1Y1M")))
        self.assertFalse(self.getLib().durationEqual(self.makeDuration("P1Y1M"), None))

        self.assertTrue(self.getLib().durationEqual(self.makeDuration("P1Y1M"), self.makeDuration("P1Y1M")))
        self.assertTrue(self.getLib().durationEqual(self.makeDuration("P1Y1M"), self.makeDuration("P0Y13M")))
        self.assertFalse(self.getLib().durationEqual(self.makeDuration("P1Y1M"), self.makeDuration("P1Y2M")))

        self.assertTrue(self.getLib().durationEqual(self.makeDuration("P1DT1H"), self.makeDuration("P1DT1H")))
        self.assertTrue(self.getLib().durationEqual(self.makeDuration("P1DT1H"), self.makeDuration("P0DT25H")))
        self.assertFalse(self.getLib().durationEqual(self.makeDuration("P1DT1H"), self.makeDuration("P1DT2H")))

    def testDurationNotEqual(self):
        self.assertFalse(self.getLib().durationNotEqual(None, None))
        self.assertTrue(self.getLib().durationNotEqual(None, self.makeDuration("P1Y1M")))
        self.assertTrue(self.getLib().durationNotEqual(self.makeDuration("P1Y1M"), None))

        self.assertFalse(self.getLib().durationNotEqual(self.makeDuration("P1Y1M"), self.makeDuration("P1Y1M")))
        self.assertFalse(self.getLib().durationNotEqual(self.makeDuration("P1Y1M"), self.makeDuration("P0Y13M")))
        self.assertTrue(self.getLib().durationNotEqual(self.makeDuration("P1Y1M"), self.makeDuration("P1Y2M")))

        self.assertFalse(self.getLib().durationNotEqual(self.makeDuration("P1DT1H"), self.makeDuration("P1DT1H")))
        self.assertFalse(self.getLib().durationNotEqual(self.makeDuration("P1DT1H"), self.makeDuration("P0DT25H")))
        self.assertTrue(self.getLib().durationNotEqual(self.makeDuration("P1DT1H"), self.makeDuration("P1DT2H")))

    def testDurationLessThan(self):
        self.assertIsNone(self.getLib().durationLessThan(None, None))
        self.assertIsNone(self.getLib().durationLessThan(None, self.makeDuration("P1Y1M")))
        self.assertIsNone(self.getLib().durationLessThan(self.makeDuration("P1Y1M"), None))

        self.assertFalse(self.getLib().durationLessThan(self.makeDuration("P1Y1M"), self.makeDuration("P1Y1M")))
        self.assertFalse(self.getLib().durationLessThan(self.makeDuration("P1Y1M"), self.makeDuration("P0Y13M")))
        self.assertTrue(self.getLib().durationLessThan(self.makeDuration("P1Y1M"), self.makeDuration("P1Y2M")))

        self.assertFalse(self.getLib().durationLessThan(self.makeDuration("P1DT1H"), self.makeDuration("P1DT1H")))
        self.assertFalse(self.getLib().durationLessThan(self.makeDuration("P1DT1H"), self.makeDuration("P0DT25H")))
        self.assertTrue(self.getLib().durationLessThan(self.makeDuration("P1DT1H"), self.makeDuration("P1DT2H")))

    def testDurationGreaterThan(self):
        self.assertIsNone(self.getLib().durationGreaterThan(None, None))
        self.assertIsNone(self.getLib().durationGreaterThan(None, self.makeDuration("P1Y1M")))
        self.assertIsNone(self.getLib().durationGreaterThan(self.makeDuration("P1Y1M"), None))

        self.assertFalse(self.getLib().durationGreaterThan(self.makeDuration("P1Y1M"), self.makeDuration("P1Y1M")))
        self.assertFalse(self.getLib().durationGreaterThan(self.makeDuration("P1Y1M"), self.makeDuration("P0Y13M")))
        self.assertFalse(self.getLib().durationGreaterThan(self.makeDuration("P1Y1M"), self.makeDuration("P1Y2M")))

        self.assertFalse(self.getLib().durationGreaterThan(self.makeDuration("P1DT1H"), self.makeDuration("P1DT1H")))
        self.assertFalse(self.getLib().durationGreaterThan(self.makeDuration("P1DT1H"), self.makeDuration("P0DT25H")))
        self.assertFalse(self.getLib().durationGreaterThan(self.makeDuration("P1DT1H"), self.makeDuration("P1DT2H")))

    def testDurationLessEqualThan(self):
        self.assertTrue(self.getLib().durationLessEqualThan(None, None))
        self.assertIsNone(self.getLib().durationLessEqualThan(None, self.makeDuration("P1Y1M")))
        self.assertIsNone(self.getLib().durationLessEqualThan(self.makeDuration("P1Y1M"), None))

        self.assertTrue(self.getLib().durationLessEqualThan(self.makeDuration("P1Y1M"), self.makeDuration("P1Y1M")))
        self.assertTrue(self.getLib().durationLessEqualThan(self.makeDuration("P1Y1M"), self.makeDuration("P0Y13M")))
        self.assertTrue(self.getLib().durationLessEqualThan(self.makeDuration("P1Y1M"), self.makeDuration("P1Y2M")))

        self.assertTrue(self.getLib().durationLessEqualThan(self.makeDuration("P1DT1H"), self.makeDuration("P1DT1H")))
        self.assertTrue(self.getLib().durationLessEqualThan(self.makeDuration("P1DT1H"), self.makeDuration("P0DT25H")))
        self.assertTrue(self.getLib().durationLessEqualThan(self.makeDuration("P1DT1H"), self.makeDuration("P1DT2H")))

    def testDurationGreaterEqualThan(self):
        self.assertTrue(self.getLib().durationGreaterEqualThan(None, None))
        self.assertIsNone(self.getLib().durationGreaterEqualThan(None, self.makeDuration("P1Y1M")))
        self.assertIsNone(self.getLib().durationGreaterEqualThan(self.makeDuration("P1Y1M"), None))

        self.assertTrue(self.getLib().durationGreaterEqualThan(self.makeDuration("P1Y1M"), self.makeDuration("P1Y1M")))
        self.assertTrue(self.getLib().durationGreaterEqualThan(self.makeDuration("P1Y1M"), self.makeDuration("P0Y13M")))
        self.assertFalse(self.getLib().durationGreaterEqualThan(self.makeDuration("P1Y1M"), self.makeDuration("P1Y2M")))

        self.assertTrue(self.getLib().durationGreaterEqualThan(self.makeDuration("P1DT1H"), self.makeDuration("P1DT1H")))
        self.assertTrue(self.getLib().durationGreaterEqualThan(self.makeDuration("P1DT1H"), self.makeDuration("P0DT25H")))
        self.assertFalse(self.getLib().durationGreaterEqualThan(self.makeDuration("P1DT1H"), self.makeDuration("P1DT2H")))

    def testDurationAdd(self):
        self.assertIsNone(self.getLib().durationAdd(None, None))
        self.assertIsNone(self.getLib().durationAdd(None, self.makeDuration("P1Y1M")))
        self.assertIsNone(self.getLib().durationAdd(self.makeDuration("P1Y1M"), None))

        self.assertEqual(self.makeDuration("P2Y2M"), self.getLib().durationAdd(self.makeDuration("P1Y1M"), self.makeDuration("P1Y1M")))
        self.assertEqual(self.makeDuration("P2Y3M"), self.getLib().durationAdd(self.makeDuration("P1Y1M"), self.makeDuration("P1Y2M")))

        self.assertEqual(self.makeDuration("P2DT2H"), self.getLib().durationAdd(self.makeDuration("P1DT1H"), self.makeDuration("P1DT1H")))
        self.assertEqual(self.makeDuration("P2DT3H"), self.getLib().durationAdd(self.makeDuration("P1DT1H"), self.makeDuration("P1DT2H")))

        self.assertEqual(self.makeDuration("P380DT8H59M13S"), self.getLib().durationAdd(
            self.makeDuration("P13DT2H14S"), self.getLib().dateTimeSubtract(self.makeDateAndTime("2016-12-24T23:59:00-08:00"), self.makeDateAndTime("2015-12-24T00:00:01-01:00"))))

    def testDurationSubtract(self):
        self.assertIsNone(self.getLib().durationSubtract(None, None))
        self.assertIsNone(self.getLib().durationSubtract(None, self.makeDuration("P1Y1M")))
        self.assertIsNone(self.getLib().durationSubtract(self.makeDuration("P1Y1M"), None))

        self.assertEqual(self.makeDuration("P0Y0M"), self.getLib().durationSubtract(self.makeDuration("P1Y1M"), self.makeDuration("P1Y1M")))
        self.assertEqual(self.makeDuration("-P0Y1M"), self.getLib().durationSubtract(self.makeDuration("P1Y1M"), self.makeDuration("P1Y2M")))

        self.assertEqual(self.makeDuration("P0DT0H"), self.getLib().durationSubtract(self.makeDuration("P1DT1H"), self.makeDuration("P1DT1H")))
        self.assertEqual(self.makeDuration("-P0DT1H"), self.getLib().durationSubtract(self.makeDuration("P1DT1H"), self.makeDuration("P1DT2H")))

    def testDurationMultiply(self):
        self.assertIsNone(self.getLib().durationMultiplyNumber(None, None))
        self.assertIsNone(self.getLib().durationMultiplyNumber(None, self.makeNumber("2")))
        self.assertIsNone(self.getLib().durationMultiplyNumber(self.makeDuration("P1Y1M"), None))

        self.assertEqual(self.makeDuration("P2Y2M"), self.getLib().durationMultiplyNumber(self.makeDuration("P1Y1M"), self.makeNumber("2")))
        self.assertEqual(self.makeDuration("-P2Y2M"), self.getLib().durationMultiplyNumber(self.makeDuration("P1Y1M"), self.makeNumber("-2")))

        self.assertEqual(self.makeDuration("P2DT2H"), self.getLib().durationMultiplyNumber(self.makeDuration("P1DT1H"), self.makeNumber("2")))
        self.assertEqual(self.makeDuration("-P2DT2H"), self.getLib().durationMultiplyNumber(self.makeDuration("P1DT1H"), self.makeNumber("-2")))

    def testDurationDivide(self):
        self.assertIsNone(self.getLib().durationDivideNumber(None, None))
        self.assertIsNone(self.getLib().durationDivideNumber(None, self.makeNumber("2")))
        self.assertIsNone(self.getLib().durationDivideNumber(self.makeDuration("P1Y1M"), None))

        self.assertEqual(self.makeDuration("P0Y6M"), self.getLib().durationDivideNumber(self.makeDuration("P1Y1M"), self.makeNumber("2")))
        self.assertEqual(self.makeDuration("P1Y1M"), self.getLib().durationDivideNumber(self.makeDuration("P2Y2M"), self.makeNumber("2")))

        self.assertEqual(self.makeDuration("P0DT12H30M"), self.getLib().durationDivideNumber(self.makeDuration("P1DT1H"), self.makeNumber("2")))
        self.assertEqual(self.makeDuration("P1DT1H"), self.getLib().durationDivideNumber(self.makeDuration("P2DT2H"), self.makeNumber("2")))

    #
    # List operators
    #
    def testIsList(self):
        self.assertFalse(self.getLib().isList(None))
        self.assertFalse(self.getLib().isList(self.getLib().number("1")))
        self.assertFalse(self.getLib().isList("abc"))
        self.assertFalse(self.getLib().isList(True))
        self.assertFalse(self.getLib().isList(self.getLib().date("2020-01-01")))
        self.assertFalse(self.getLib().isList(self.getLib().time("12:00:00")))
        self.assertFalse(self.getLib().isList(self.getLib().dateAndTime("2020-01-01T12:00:00")))
        self.assertFalse(self.getLib().isList(self.getLib().duration("P1Y1M")))
        self.assertTrue(self.getLib().isList(self.getLib().asList("a")))
        self.assertFalse(self.getLib().isList(Context()))
        self.assertFalse(self.getLib().isList(Range(True, 0, True, 1)))

    def testListIs(self):
        self.assertTrue(self.getLib().listIs(None, None))
        self.assertFalse(self.getLib().listIs(self.makeList("a"), None))
        self.assertFalse(self.getLib().listIs(None, self.makeList("a")))

        self.assertFalse(self.getLib().listIs(self.makeList("a"), self.makeList("b")))
        self.assertTrue(self.getLib().listIs(self.makeList("a"), self.makeList("a")))

    def testListEqual(self):
        self.assertTrue(self.getLib().listEqual(None, None))
        self.assertFalse(self.getLib().listEqual(self.makeList("a"), None))
        self.assertFalse(self.getLib().listEqual(None, self.makeList("a")))

        self.assertFalse(self.getLib().listEqual(self.makeList("a"), self.makeList("b")))
        self.assertTrue(self.getLib().listEqual(self.makeList("a"), self.makeList("a")))

    def testListNotEqual(self):
        self.assertFalse(self.getLib().listNotEqual(None, None))
        self.assertTrue(self.getLib().listNotEqual(self.makeList("a"), None))
        self.assertTrue(self.getLib().listNotEqual(None, self.makeList("a")))

        self.assertTrue(self.getLib().listNotEqual(self.makeList("a"), self.makeList("b")))
        self.assertFalse(self.getLib().listNotEqual(self.makeList("a"), self.makeList("a")))

    #
    # Context operators
    #
    def testIsContext(self):
        self.assertFalse(self.getLib().isContext(None))
        self.assertFalse(self.getLib().isContext(self.getLib().number("1")))
        self.assertFalse(self.getLib().isContext("abc"))
        self.assertFalse(self.getLib().isContext(True))
        self.assertFalse(self.getLib().isContext(self.getLib().date("2020-01-01")))
        self.assertFalse(self.getLib().isContext(self.getLib().time("12:00:00")))
        self.assertFalse(self.getLib().isContext(self.getLib().dateAndTime("2020-01-01T12:00:00")))
        self.assertFalse(self.getLib().isContext(self.getLib().duration("P1Y1M")))
        self.assertFalse(self.getLib().isContext(self.getLib().asList("a")))
        self.assertTrue(self.getLib().isContext(Context()))
        self.assertFalse(self.getLib().isContext(Range(True, 0, True, 1)))

    def testContextValue(self):
        c1 = Context()

        self.assertIsNone(self.getLib().contextValue(None))
        self.assertIsNone(self.getLib().contextValue(self.getLib().number("123")))
        self.assertEqual(c1, self.getLib().contextValue(c1))

    def testContextIs(self):
        c1 = Context().add("m", "a")
        c2 = Context().add("m", "b")
        c3 = Context().add("m", "a")

        self.assertTrue(self.getLib().contextIs(None, None))
        self.assertFalse(self.getLib().contextIs(c1, None))
        self.assertFalse(self.getLib().contextIs(None, c1))

        self.assertTrue(self.getLib().contextIs(c1, c1))
        self.assertFalse(self.getLib().contextIs(c1, c2))
        self.assertTrue(self.getLib().contextIs(c1, c3))

    def testContextEqual(self):
        c1 = Context().add("m", "a")
        c2 = Context().add("m", "b")
        c3 = Context().add("m", "a")

        self.assertTrue(self.getLib().contextEqual(None, None))
        self.assertFalse(self.getLib().contextEqual(c1, None))
        self.assertFalse(self.getLib().contextEqual(None, c1))

        self.assertFalse(self.getLib().contextEqual(c1, c2))
        self.assertTrue(self.getLib().contextEqual(c1, c3))

    def testContextNotEqual(self):
        c1 = Context().add("m", "a")
        c2 = Context().add("m", "b")
        c3 = Context().add("m", "a")

        self.assertFalse(self.getLib().contextNotEqual(None, None))
        self.assertTrue(self.getLib().contextNotEqual(c1, None))
        self.assertTrue(self.getLib().contextNotEqual(None, c1))

        self.assertTrue(self.getLib().contextNotEqual(c1, c2))
        self.assertFalse(self.getLib().contextNotEqual(c1, c3))

    #
    # Range operators
    #
    def testIsRange(self):
        self.assertFalse(self.getLib().isRange(None))
        self.assertFalse(self.getLib().isRange(self.getLib().number("1")))
        self.assertFalse(self.getLib().isRange("abc"))
        self.assertFalse(self.getLib().isRange(True))
        self.assertFalse(self.getLib().isRange(self.getLib().date("2020-01-01")))
        self.assertFalse(self.getLib().isRange(self.getLib().time("12:00:00")))
        self.assertFalse(self.getLib().isRange(self.getLib().dateAndTime("2020-01-01T12:00:00")))
        self.assertFalse(self.getLib().isRange(self.getLib().duration("P1Y1M")))
        self.assertFalse(self.getLib().isRange(self.getLib().asList("a")))
        self.assertFalse(self.getLib().isRange(Context()))
        self.assertTrue(self.getLib().isRange(Range(True, 0, True, 1)))

    def testRangeValue(self):
        r1 = Range(True, self.getLib().number("1"), True, self.getLib().number("2"))

        self.assertIsNone(self.getLib().rangeValue(None))
        self.assertEqual(r1, self.getLib().rangeValue(r1))

    def testRangeIs(self):
        r1 = Range(True, self.getLib().number("1"), True, self.getLib().number("2"))
        r2 = Range(True, self.getLib().number("1"), True, self.getLib().number("3"))
        r3 = Range(True, self.getLib().number("1"), True, self.getLib().number("2"))

        self.assertTrue(self.getLib().rangeIs(None, None))
        self.assertFalse(self.getLib().rangeIs(r1, None))
        self.assertFalse(self.getLib().rangeIs(None, r1))

        self.assertTrue(self.getLib().rangeIs(r1, r1))
        self.assertFalse(self.getLib().rangeIs(r1, r2))
        self.assertTrue(self.getLib().rangeIs(r1, r3))

    def testRangeEqual(self):
        r1 = Range(True, self.getLib().number("1"), True, self.getLib().number("2"))
        r2 = Range(True, self.getLib().number("1"), True, self.getLib().number("3"))
        r3 = Range(True, self.getLib().number("1"), True, self.getLib().number("2"))

        self.assertTrue(self.getLib().rangeEqual(None, None))
        self.assertFalse(self.getLib().rangeEqual(r1, None))
        self.assertFalse(self.getLib().rangeEqual(None, r1))

        self.assertFalse(self.getLib().rangeEqual(r1, r2))
        self.assertTrue(self.getLib().rangeEqual(r1, r3))

    def testRangeNotEqual(self):
        r1 = Range(True, self.getLib().number("1"), True, self.getLib().number("2"))
        r2 = Range(True, self.getLib().number("1"), True, self.getLib().number("3"))
        r3 = Range(True, self.getLib().number("1"), True, self.getLib().number("2"))

        self.assertFalse(self.getLib().rangeNotEqual(None, None))
        self.assertTrue(self.getLib().rangeNotEqual(r1, None))
        self.assertTrue(self.getLib().rangeNotEqual(None, r1))

        self.assertTrue(self.getLib().rangeNotEqual(r1, r2))
        self.assertFalse(self.getLib().rangeNotEqual(r1, r3))

    #
    # Function operators
    #
    def testIsFunction(self):
        self.assertTrue(self.getLib().isFunction(None))
        self.assertFalse(self.getLib().isFunction(self.getLib().number("1")))
        self.assertFalse(self.getLib().isFunction("abc"))
        self.assertFalse(self.getLib().isFunction(True))
        self.assertFalse(self.getLib().isFunction(self.getLib().date("2020-01-01")))
        self.assertFalse(self.getLib().isFunction(self.getLib().time("12:00:00")))
        self.assertFalse(self.getLib().isFunction(self.getLib().dateAndTime("2020-01-01T12:00:00")))
        self.assertFalse(self.getLib().isFunction(self.getLib().dateAndTime("2020-01-01T12:00:00")))
        self.assertFalse(self.getLib().isFunction(self.getLib().duration("P1Y1M")))
        self.assertFalse(self.getLib().isFunction(self.getLib().asList("a")))
        self.assertFalse(self.getLib().isFunction(Context()))
        self.assertFalse(self.getLib().isFunction(Range(True, 0, True, 1)))

    def testFunctionValue(self):
        self.assertIsNone(self.getLib().functionValue(None))
        self.assertIsNone(self.getLib().functionValue("a"))

    def testFunctionIs(self):
        self.assertTrue(self.getLib().functionIs(None, None))
        self.assertIsNone(self.getLib().functionIs("a", None))
        self.assertIsNone(self.getLib().functionIs(None, "b"))

        self.assertIsNone(self.getLib().functionIs("a", "b"))
        self.assertIsNone(self.getLib().functionIs("b", "b"))

    def testFunctionEqual(self):
        self.assertTrue(self.getLib().functionEqual(None, None))
        self.assertIsNone(self.getLib().functionEqual("a", None))
        self.assertIsNone(self.getLib().functionEqual(None, "b"))

        self.assertIsNone(self.getLib().functionEqual("a", "b"))
        self.assertIsNone(self.getLib().functionEqual("b", "b"))

    def testFunctionNotEqual(self):
        self.assertFalse(self.getLib().functionNotEqual(None, None))
        self.assertIsNone(self.getLib().functionNotEqual("a", None))
        self.assertIsNone(self.getLib().functionNotEqual(None, "b"))

        self.assertIsNone(self.getLib().functionNotEqual("a", "b"))
        self.assertIsNone(self.getLib().functionNotEqual("b", "b"))

    #
    # Common
    #
    def getLib(self) -> BaseStandardFEELLib:
        return DefaultStandardFEELLib()

    def makeNumber(self, literal: Any) -> Decimal:
        return self.getLib().number(literal)

    def makeNumberList(self, *numbers):
        return [self.makeNumber(str(n)) for n in numbers]

    def makeDate(self, literal: str) -> date:
        return self.getLib().date(literal)

    def makeTime(self, literal: str) -> time:
        return self.getLib().time(literal)

    def makeDateAndTime(self, literal: str) -> datetime:
        return self.getLib().dateAndTime(literal)

    def makeDuration(self, literal: str) -> Duration:
        return self.getLib().duration(literal)

    @staticmethod
    def makeList(*args) -> List[Any]:
        return [x for x in args]

    def assertEqualsNumber(self, expected: Any, actual: Any, delta: float = None):
        if isinstance(expected, str):
            expected = self.makeNumber(expected)

        self.assertAlmostEqual(expected, actual, delta=delta)

    def assertEqualsDate(self, expected: Any, actual: Optional[date]):
        if isinstance(expected, str):
            expected = self.makeDate(expected)

        self.assertEqual(expected, actual)

    def assertEqualsTime(self, expected: Any, actual: Optional[time]):
        if isinstance(expected, str):
            expected = self.makeTime(expected)

        self.assertEqual(expected, actual)

    def assertEqualsDateTime(self, expected: Any, actual: Optional[datetime]):
        if isinstance(actual, datetime):
            expected = self.makeDateAndTime(expected)
        elif isinstance(actual, date):
            expected = self.makeDate(expected)
        elif isinstance(actual, time):
            expected = self.makeTime(expected)
        elif isinstance(actual, Duration):
            expected = self.makeDuration(expected)
        elif isinstance(actual, timedelta):
            expected = self.makeDuration(expected)

        self.assertEqual(expected, actual)

    def assertEqualsList(self, expected: Any, actual: Any):
        Assert().assertEquals(expected, str(actual))

    def assertEqualsUnicodeString(self, expected: Any, actual: Any):
        Assert().assertEquals(expected, actual)



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from typing import Any, List, Optional

from jdmn.feel.lib.type.BaseType import BaseType
from jdmn.feel.lib.type.bool.DefaultBooleanType import DefaultBooleanType


class DefaultListType(BaseType):
    def __init__(self):
        BaseType.__init__(self)
        self.booleanType = DefaultBooleanType()

    def isList(self, value: Any) -> bool:
        return isinstance(value, List)

    def listValue(self, value: Optional[List[Any]]) -> Optional[List[Any]]:
        return value

    def listIs(self, list1: Optional[List[Any]], list2: Optional[List[Any]]) -> Optional[bool]:
        return self.listEqual(list1, list2)

    def listEqual(self, list1: Optional[List[Any]], list2: Optional[List[Any]]) -> Optional[bool]:
        if list1 is None and list2 is None:
            return True
        elif list1 is None:
            return False
        elif list2 is None:
            return False
        else:
            return self.isList(list1) and self.isList(list2) and list1 == list2

    def listNotEqual(self, list1: Optional[List[Any]], list2: Optional[List[Any]]) -> Optional[bool]:
        return self.booleanType.booleanNot(self.listEqual(list1, list2))



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use self file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#

#
# Lazily evaluate the supplied operations to reduce unnecessary computation
#
# @param <T> the type of value stored
#
#
import logging
import typing


class LazyEval:
    LOGGER = logging.getLogger(__name__)

    def __init__(self, supplier: typing.Callable):
        self.supplier = supplier
        self.isValueSet = False

    def getOrCompute(self) -> typing.Any:
        return self.value if self.isValueSet else self.compute()

    def compute(self) -> typing.Any:
        self.LOGGER.info("Trigger lazy evaluation")

        self.isValueSet = True
        self.value = self.supplier()
        return self.value



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
import datetime
from typing import Union

from jdmn.feel.lib.Types import BOOLEAN
from jdmn.feel.lib.type.RelationalComparator import RelationalComparator

comparable = Union[float, str, datetime.date, datetime.time, datetime.datetime]


class ComparableComparator(RelationalComparator):
    def __init__(self):
        super().__init__()

    def compare(self, first: comparable, second: comparable) -> int:
        return self.compareTo(first, second)

    def equalTo(self, first: comparable, second: comparable) -> BOOLEAN:
        return self.applyOperator(first, second, [
            lambda: True,
            lambda: False,
            lambda: False,
            lambda: type(first) == type(second) and self.compareTo(first, second) == 0
        ])

    def lessThan(self, first: comparable, second: comparable) -> BOOLEAN:
        return self.applyOperator(first, second, [
            lambda: None,
            lambda: None,
            lambda: None,
            lambda: type(first) == type(second) and self.compareTo(first, second) < 0
        ])

    def compareTo(self, first: comparable, second: comparable) -> int:
        pass



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from datetime import datetime, date, time, timedelta
from decimal import Decimal

from isodate import Duration

from jdmn.feel.lib.Types import POINT_RANGE_UNION, BOOLEAN, RANGE, COMPARABLE
from jdmn.feel.lib.type.bool.DefaultBooleanType import DefaultBooleanType
from jdmn.feel.lib.type.numeric.NumericComparator import NumericComparator
from jdmn.feel.lib.type.string.StringComparator import StringComparator
from jdmn.feel.lib.type.time.DefaultDateComparator import DefaultDateComparator
from jdmn.feel.lib.type.time.DefaultDateTimeComparator import DefaultDateTimeComparator
from jdmn.feel.lib.type.time.DefaultDurationComparator import DefaultDurationComparator
from jdmn.feel.lib.type.time.DefaultTimeComparator import DefaultTimeComparator
from jdmn.runtime.Range import Range

COMPARATOR_MAP = {}
COMPARATOR_MAP[Decimal] = NumericComparator()
COMPARATOR_MAP[str] = StringComparator()
COMPARATOR_MAP[date] = DefaultDateComparator()
COMPARATOR_MAP[time] = DefaultTimeComparator()
COMPARATOR_MAP[datetime] = DefaultDateTimeComparator()
COMPARATOR_MAP[Duration] = DefaultDurationComparator()
COMPARATOR_MAP[timedelta] = DefaultDurationComparator()


class DefaultRangeLib:
    BOOLEAN_TYPE = DefaultBooleanType()

    def before(self, arg1: POINT_RANGE_UNION, arg2: POINT_RANGE_UNION) -> BOOLEAN:
        if self.checkArguments(arg1, arg2):
            return None

        #
        # DMN 1.3 spec
        #
        if self.isPoint(arg1) and self.isPoint(arg2):
            # point1 < point2
            rc = self.resolveComparator(arg1)
            return None if rc is None else rc.lessThan(arg1, arg2)
        elif self.isPoint(arg1) and self.isRange(arg2):
            # point < range.start
            # or (point = range.start and not(range.start included) )
            point: COMPARABLE = arg1
            range_: RANGE = arg2
            start = range_.getStart()
            rc = self.resolveComparator(point)
            return None if rc is None else self.BOOLEAN_TYPE.booleanOr(
                rc.lessThan(point, start),
                self.BOOLEAN_TYPE.booleanAnd(rc.equalTo(point, start), not range_.isStartIncluded())
            )
        elif self.isRange(arg1) and self.isPoint(arg2):
            # range.end < point
            # or (range.end = point and not (range.end included) )
            range_ = arg1
            point = arg2
            end = range_.getEnd()
            rc = self.resolveComparator(point)
            return None if rc is None else self.BOOLEAN_TYPE.booleanOr(
                rc.lessThan(end, point),
                self.BOOLEAN_TYPE.booleanAnd(rc.equalTo(end, point), not range_.isEndIncluded())
            )
        else:
            # range1.end < range2.start
            # or ((not (range1.end included) or not (range2.start included)) and range1.end = range2.start)
            range1 = arg1
            range2 = arg2
            start1 = range1.getStart()
            end1 = range1.getEnd()
            start2 = range2.getStart()
            rc = self.resolveComparator(start1)
            return None if rc is None else self.BOOLEAN_TYPE.booleanOr(
                rc.lessThan(end1, start2),
                self.BOOLEAN_TYPE.booleanAnd(
                    self.BOOLEAN_TYPE.booleanOr(not range1.isEndIncluded(), not range2.isStartIncluded()),
                    rc.equalTo(end1, start2)
                )
            )

    def after(self, arg1: POINT_RANGE_UNION, arg2: POINT_RANGE_UNION):
        if self.checkArguments(arg1, arg2):
            return None

        #
        # DMN 1.3 spec
        #
        if self.isPoint(arg1) and self.isPoint(arg2):
            # point1 > point2
            rc = self.resolveComparator(arg1)
            return None if rc is None else rc.greaterThan(arg1, arg2)
        elif self.isPoint(arg1) and self.isRange(arg2):
            # point > range.end
            # or (point = range.end and not(range.end included) )
            point = arg1
            range_ = arg2
            end = range_.getEnd()
            rc = self.resolveComparator(point)
            return None if rc is None else self.BOOLEAN_TYPE.booleanOr(
                    rc.greaterThan(point, end),
                    self.BOOLEAN_TYPE.booleanAnd(
                         rc.equalTo(point, end),
                         not range_.isEndIncluded()
                    )
            )
        elif self.isRange(arg1) and self.isPoint(arg2):
            # range.start > point
            # or (range.start = point and not(range.start included) )
            range_ = arg1
            point = arg2
            start = range_.getStart()
            rc = self.resolveComparator(start)
            return None if rc is None else self.BOOLEAN_TYPE.booleanOr(
                    rc.greaterThan(start, point),
                    self.BOOLEAN_TYPE.booleanAnd(
                            rc.equalTo(start, point),
                            not range_.isStartIncluded()
                    )
            )
        else:
            # range1.start > range2.end
            # or ((not (range1.start included) or not (range2.end included)) and range1.start = range2.end)
            range1 = arg1
            range2 = arg2
            start1 = range1.getStart()
            end2 = range2.getEnd()
            rc = self.resolveComparator(start1)
            return None if rc is None else self.BOOLEAN_TYPE.booleanOr(
                rc.greaterThan(start1, end2),
                self.BOOLEAN_TYPE.booleanAnd(
                    self.BOOLEAN_TYPE.booleanOr(not range1.isStartIncluded(), not range2.isEndIncluded()),
                    rc.equalTo(start1, end2)
                )
            )

    def meets(self, range1: RANGE, range2: RANGE) -> BOOLEAN:
        if self.checkArguments(range1, range2):
            return None
        if not self.isRange(range1) or not self.isRange(range2):
            return None

        #
        # DMN 1.3 spec
        # range1.end included
        # and range2.start included
        # and range1.end = range2.start
        #
        start1 = range1.getStart()
        end1 = range1.getEnd()
        start2 = range2.getStart()
        rc = self.resolveComparator(start1)
        return None if rc is None else self.BOOLEAN_TYPE.booleanAnd(
                range1.isEndIncluded(),
                range2.isStartIncluded(),
                rc.equalTo(end1, start2)
        )

    def metBy(self, range1: RANGE, range2: RANGE) -> BOOLEAN:
        if self.checkArguments(range1, range2):
            return None
        if not self.isRange(range1) or not self.isRange(range2):
            return None

        #
        # DMN 1.3 spec
        # range1.start included
        # and range2.end included
        # and range1.start = range2.end
        #
        start1 = range1.getStart()
        end2 = range2.getEnd()
        rc = self.resolveComparator(start1)
        return None if rc is None else self.BOOLEAN_TYPE.booleanAnd(
                range1.isStartIncluded(),
                range2.isEndIncluded(),
                rc.equalTo(start1, end2)
        )

    def overlaps(self, range1: RANGE, range2: RANGE) -> BOOLEAN:
        if self.checkArguments(range1, range2):
            return None
        if not self.isRange(range1) or not self.isRange(range2):
            return None

        #
        # CQL spec
        #
        # The overlaps operator returns true if the first interval overlaps the second.
        # More precisely,
        #      if the starting or ending point of either interval is in the other,
        #      or if the ending point of the first interval is greater than or equal to the starting point of the second interval,
        #      and the starting point of the first interval is less than or equal to the ending point of the second interval.
        #
        # DMN 1.3 spec
        # (range1.end > range2.start or (range1.end = range2.start and (range1.end included or range2.end included)))
        # and (range1.start < range2.end or (range1.start = range2.end and range1.start included and range2.end included))
        #
        # Corrected:
        # overlaps before(range1, range2)
        # or overlaps after(range1, range2)
        # or includes(range1, range2)
        # or includes(range2, range1)
        #
        return self.BOOLEAN_TYPE.booleanOr(
            self.overlapsBefore(range1, range2),
            self.overlapsAfter(range1, range2),
            self.includes(range1, range2),
            self.includes(range2, range1)
        )

    def overlapsBefore(self, range1: RANGE, range2: RANGE) -> BOOLEAN:
        if self.checkArguments(range1, range2):
            return None
        if not self.isRange(range1) or not self.isRange(range2):
            return None

        #
        # CQL spec:
        # The operator overlaps before returns true if the first interval overlaps the second and starts before it,
        # while the overlaps after operator returns true if the first interval overlaps the second and ends after it.
        #

        #
        # DMN 1.3 spec
        # range1 starts before range2
        # and range1.end is in range2
        #
        # (range1.start < range2.start or (range1.start = range2.start and range1.start included and range2.start included))
        # and (range1.end > range2.start or (range1.end = range2.start and range1.end included and range2.start included))
        # and (range1.end < range2.end  or (range1.end = range2.end and (not(range1.end included) or range2.end included )))
        #

        # Corrected to
        # (range1.start < range2.start or (range1.start = range2.start and (range1.start included and not(range2.start included))))
        # and (range1.end > range2.start or (range1.end = range2.start and range1.end included and range2.start included))
        # and (range1.end < range2.end  or (range1.end = range2.end and (not(range1.end included) or range2.end included )))
        start1 = range1.getStart()
        end1 = range1.getEnd()
        start2 = range2.getStart()
        end2 = range2.getEnd()
        rc = self.resolveComparator(start1)
        if rc is None:
            return None

        range1StartsBeforeRange2 = self.BOOLEAN_TYPE.booleanOr(
                rc.lessThan(start1, start2),
                self.BOOLEAN_TYPE.booleanAnd(
                        rc.equalTo(start1, start2),
                        range1.isStartIncluded(),
                        not range2.isStartIncluded()
                )
        )
        range1EndIsInRange2 = self.BOOLEAN_TYPE.booleanAnd(
                self.BOOLEAN_TYPE.booleanOr(
                    rc.greaterThan(end1, start2),
                    self.BOOLEAN_TYPE.booleanAnd(
                        rc.equalTo(end1, start2),
                        range1.isEndIncluded(),
                        range2.isStartIncluded()
                    )
                ),
                self.BOOLEAN_TYPE.booleanOr(
                        rc.lessThan(end1, end2),
                        self.BOOLEAN_TYPE.booleanAnd(
                                rc.equalTo(end1, end2),
                                self.BOOLEAN_TYPE.booleanOr(
                                        not range1.isEndIncluded(),
                                        range2.isEndIncluded()
                                )
                        )
                )
        )
        return self.BOOLEAN_TYPE.booleanAnd(
                range1StartsBeforeRange2,
                range1EndIsInRange2
        )

    def overlapsAfter(self, range1: RANGE, range2: RANGE) -> BOOLEAN:
        if self.checkArguments(range1, range2):
            return None
        if not self.isRange(range1) or not self.isRange(range2):
            return None

        #
        # DMN 1.3 spec
        # range2 starts before range1
        # range2.end is in range1
        #
        # (range2.start < range1.start or (range2.start = range1.start and range2.start included and not( range1.start included)))
        # and (range2.end > range1.start or (range2.end = range1.start and range2.end included and range1.start included ))
        # and (range2.end < range1.end or (range2.end = range1.end and (not(range2.end included) or range1.end included)))
        #
        start1 = range1.getStart()
        end1 = range1.getEnd()
        start2 = range2.getStart()
        end2 = range2.getEnd()
        rc = self.resolveComparator(start1)
        if rc is None:
            return None

        range2StartsBeforeRange1 = self.BOOLEAN_TYPE.booleanOr(
                rc.lessThan(start2, start1),
                self.BOOLEAN_TYPE.booleanAnd(rc.equalTo(start2, start1), range2.isStartIncluded(), not range1.isStartIncluded())
        )
        range2EndIsInRange1 = self.BOOLEAN_TYPE.booleanAnd(
                self.BOOLEAN_TYPE.booleanOr(
                    rc.greaterThan(end2, start1),
                    self.BOOLEAN_TYPE.booleanAnd(
                            rc.equalTo(end2, start1),
                            range2.isEndIncluded(),
                            range1.isStartIncluded()
                    )
                ),
                self.BOOLEAN_TYPE.booleanOr(
                        rc.lessThan(end2, end1),
                        self.BOOLEAN_TYPE.booleanAnd(
                                rc.equalTo(end2, end1),
                                self.BOOLEAN_TYPE.booleanOr(
                                        not range2.isEndIncluded(),
                                        range1.isEndIncluded()
                                )
                        )
                )
        )
        return self.BOOLEAN_TYPE.booleanAnd(
                range2StartsBeforeRange1,
                range2EndIsInRange1
        )

    def finishes(self, arg1: POINT_RANGE_UNION, arg2: RANGE) -> BOOLEAN:
        if self.checkArguments(arg1, arg2):
            return None

        #
        # DMN 1.3 spec
        #
        if self.isPoint(arg1) and self.isRange(arg2):
            point = arg1
            range_ = arg2
            # range.end included
            # and range.end = point
            end = range_.getEnd()
            rc = self.resolveComparator(point)
            return None if rc is None else self.BOOLEAN_TYPE.booleanAnd(
                range_.isEndIncluded(),
                rc.equalTo(end, point)
            )
        elif self.isRange(arg1) and self.isRange(arg2):
            range1 = arg1
            range2 = arg2
            # range1.end
            included = range2.end
            included
            # and range1.end = range2.end
            # and (range1.start > range2.start or (range1.start = range2.start and ( not (range1.start included) or range2.start included)))
            #
            start1 = range1.getStart()
            end1 = range1.getEnd()
            start2 = range2.getStart()
            end2 = range2.getEnd()
            rc = self.resolveComparator(start1)
            return None if rc is None else self.BOOLEAN_TYPE.booleanAnd(
                range1.isEndIncluded() == range2.isEndIncluded(),
                rc.equalTo(end1, end2),
                self.BOOLEAN_TYPE.booleanOr(
                    rc.greaterThan(start1, start2),
                    self.BOOLEAN_TYPE.booleanAnd(
                        rc.equalTo(start1, start2),
                        self.BOOLEAN_TYPE.booleanOr(not range1.isStartIncluded(), range2.isStartIncluded())
                    )
                )
            )
        else:
            return None

    def finishedBy(self, arg1: RANGE, arg2: POINT_RANGE_UNION) -> BOOLEAN:
        if self.checkArguments(arg1, arg2):
            return None

        #
        # DMN 1.3 spec
        #
        if self.isRange(arg1) and self.isPoint(arg2):
            range_ = arg1
            point = arg2
            # range.end included
            # and range.end = point
            start = range_.getStart()
            end = range_.getEnd()
            rc = self.resolveComparator(start)
            return None if rc is None else self.BOOLEAN_TYPE.booleanAnd(
                range_.isEndIncluded(),
                rc.equalTo(end, point)
            )
        elif self.isRange(arg1) and self.isRange(arg2):
            range1 = arg1
            range2 = arg2
            # range1.end included = range2.end included
            # and range1.end = range2.end
            # and (range1.start < range2.start or (range1.start = range2.start and (range1.start included or not (range2.start included))))
            start1 = range1.getStart()
            end1 = range1.getEnd()
            start2 = range2.getStart()
            end2 = range2.getEnd()
            rc = self.resolveComparator(start1)
            return None if rc is None else self.BOOLEAN_TYPE.booleanAnd(
                range1.isEndIncluded() == range2.isEndIncluded(),
                rc.equalTo(end1, end2),
                self.BOOLEAN_TYPE.booleanOr(
                    rc.lessThan(start1, start2),
                    self.BOOLEAN_TYPE.booleanAnd(
                        rc.equalTo(start1, start2),
                        self.BOOLEAN_TYPE.booleanOr(range1.isStartIncluded(), not range2.isStartIncluded())
                    )
                )
            )
        else:
            return None

    def includes(self, arg1: RANGE, arg2: POINT_RANGE_UNION) -> BOOLEAN:
        if self.checkArguments(arg1, arg2):
            return None

        #
        # DMN 1.3 spec
        #
        if self.isRange(arg1) and self.isPoint(arg2):
            range_ = arg1
            point = arg2
            # (range.start < point and range.end > point)
            # or (range.start = point and range.start included)
            # or (range.end = point and range.end included)
            start = range_.getStart()
            end = range_.getEnd()
            rc = self.resolveComparator(start)
            return None if rc is None else self.BOOLEAN_TYPE.booleanOr(
                self.BOOLEAN_TYPE.booleanAnd(rc.lessThan(start, point), rc.greaterThan(end, point)),
                self.BOOLEAN_TYPE.booleanAnd(rc.equalTo(start, point), range_.isStartIncluded()),
                self.BOOLEAN_TYPE.booleanAnd(rc.equalTo(end, point), range_.isEndIncluded())
            )
        elif self.isRange(arg1) and self.isRange(arg2):
            range1 = arg1
            range2 = arg2
            # (range1.start < range2.start or (range1.start = range2.start and (range1.start included or not (range2.start included))))
            # and (range1.end > range2.end or (range1.end = range2.end and (range1.end included or not (range2.end included))))
            start1 = range1.getStart()
            end1 = range1.getEnd()
            start2 = range2.getStart()
            end2 = range2.getEnd()
            rc = self.resolveComparator(start1)
            return None if rc is None else self.BOOLEAN_TYPE.booleanAnd(
                self.BOOLEAN_TYPE.booleanOr(
                    rc.lessThan(start1, start2),
                    self.BOOLEAN_TYPE.booleanAnd(
                        rc.equalTo(start1, start2),
                        self.BOOLEAN_TYPE.booleanOr(
                            range1.isStartIncluded(),
                            not range2.isStartIncluded()
                        )
                    )
                ),
                self.BOOLEAN_TYPE.booleanOr(
                    rc.greaterThan(end1, end2),
                    self.BOOLEAN_TYPE.booleanAnd(
                        rc.equalTo(end1, end2),
                        self.BOOLEAN_TYPE.booleanOr(
                            range1.isEndIncluded(),
                            not range2.isEndIncluded()
                        )
                    )
                )
            )
        else:
            return None

    def during(self, arg1: POINT_RANGE_UNION, arg2: RANGE) -> BOOLEAN:
        if self.checkArguments(arg1, arg2):
            return None

        #
        # DMN 1.3 spec
        #
        if self.isPoint(arg1) and self.isRange(arg2):
            point = arg1
            range_ = arg2
            # (range.start < point and range.end > point)
            # or (range.start = point and range.start included)
            # or (range.end = point and range.end included)
            start = range_.getStart()
            end = range_.getEnd()
            rc = self.resolveComparator(point)
            return None if rc is None else self.BOOLEAN_TYPE.booleanOr(
                self.BOOLEAN_TYPE.booleanAnd(rc.lessThan(start, point), rc.greaterThan(end, point)),
                self.BOOLEAN_TYPE.booleanAnd(rc.equalTo(start, point), range_.isStartIncluded()),
                self.BOOLEAN_TYPE.booleanAnd(rc.equalTo(end, point), range_.isEndIncluded())
            )
        elif self.isRange(arg1) and self.isRange(arg2):
            range1 = arg1
            range2 = arg2
            # (range2.start < range1.start or (range2.start = range1.start and (range2.start included or not (range1.start included))))
            # and (range2.end > range1.end or (range2.end = range1.end and (range2.end included or not (range1.end included))))
            start1 = range1.getStart()
            end1 = range1.getEnd()
            start2 = range2.getStart()
            end2 = range2.getEnd()
            rc = self.resolveComparator(start1)
            return None if rc is None else self.BOOLEAN_TYPE.booleanAnd(
                self.BOOLEAN_TYPE.booleanOr(
                    rc.lessThan(start2, start1),
                    self.BOOLEAN_TYPE.booleanAnd(
                        rc.equalTo(start2, start1),
                        self.BOOLEAN_TYPE.booleanOr(range2.isStartIncluded(), not range1.isStartIncluded())
                    )
                ),
                self.BOOLEAN_TYPE.booleanOr(
                    rc.greaterThan(end2, end1),
                    self.BOOLEAN_TYPE.booleanAnd(
                        rc.equalTo(end2, end1),
                        self.BOOLEAN_TYPE.booleanOr(range2.isEndIncluded(), not range1.isEndIncluded())
                    )
                )
            )
        else:
            return None

    def starts(self, arg1: POINT_RANGE_UNION, arg2: RANGE) -> BOOLEAN:
        if self.checkArguments(arg1, arg2):
            return None

        #
        # DMN 1.3 spec
        #
        if self.isPoint(arg1) and self.isRange(arg2):
            point = arg1
            range_ = arg2
            # range.start = point
            # and range.start included
            start = range_.getStart()
            rc = self.resolveComparator(point)
            return None if rc is None else self.BOOLEAN_TYPE.booleanAnd(
                rc.equalTo(start, point),
                range_.isStartIncluded()
            )
        elif self.isRange(arg1) and self.isRange(arg2):
            range1 = arg1
            range2 = arg2
            # range1.start = range2.start
            # and range1.start included = range2.start included
            # and (range1.end < range2.end or (range1.end = range2.end and ( not (range1.end included) or range2.end included)))
            start1 = range1.getStart()
            end1 = range1.getEnd()
            start2 = range2.getStart()
            end2 = range2.getEnd()
            rc = self.resolveComparator(start1)
            return None if rc is None else self.BOOLEAN_TYPE.booleanAnd(
                rc.equalTo(start1, start2),
                range1.isStartIncluded() == range2.isStartIncluded(),
                self.BOOLEAN_TYPE.booleanOr(
                    rc.lessThan(end1, end2),
                    self.BOOLEAN_TYPE.booleanAnd(
                        rc.equalTo(end1, end2),
                        self.BOOLEAN_TYPE.booleanOr(
                            not range1.isEndIncluded(),
                            range2.isEndIncluded()
                        )
                    )
                )
            )
        else:
            return None

    def startedBy(self, arg1: RANGE, arg2: POINT_RANGE_UNION) -> BOOLEAN:
        if self.checkArguments(arg1, arg2):
            return None

        #
        # DMN 1.3 spec
        #
        if self.isRange(arg1) and self.isPoint(arg2):
            range_ = arg1
            point = arg2
            # range.start = point
            # and range.start included
            start = range_.getStart()
            rc = self.resolveComparator(point)
            return None if rc is None else self.BOOLEAN_TYPE.booleanAnd(
                rc.equalTo(start, point),
                range_.isStartIncluded()
            )
        elif self.isRange(arg1) and self.isRange(arg2):
            range1 = arg1
            range2 = arg2
            # range1.start = range2.start
            # and range1.start included = range2.start included
            # and (range2.end < range1.end or (range2.end = range1.end and ( not (range2.end included) or range1.end included)))
            start1 = range1.getStart()
            end1 = range1.getEnd()
            start2 = range2.getStart()
            end2 = range2.getEnd()
            rc = self.resolveComparator(start1)
            return None if rc is None else self.BOOLEAN_TYPE.booleanAnd(
                rc.equalTo(start1, start2),
                range1.isStartIncluded() == range2.isStartIncluded(),
                self.BOOLEAN_TYPE.booleanOr(
                    rc.lessThan(end2, end1),
                    self.BOOLEAN_TYPE.booleanAnd(
                        rc.equalTo(end2, end1),
                        self.BOOLEAN_TYPE.booleanOr(
                            not range2.isEndIncluded(),
                            range1.isEndIncluded()
                        )
                    )
                )
            )
        else:
            return None

    def coincides(self, arg1: POINT_RANGE_UNION, arg2: POINT_RANGE_UNION) -> BOOLEAN:
        if self.checkArguments(arg1, arg2):
            return None

        #
        # DMN 1.3 spec
        #
        if self.isPoint(arg1) and self.isPoint(arg2):
            point1 = arg1
            point2 = arg2
            # point1 = point2
            rc = self.resolveComparator(point1)
            return None if rc is None else rc.equalTo(point1, point2)
        elif self.isRange(arg1) and self.isRange(arg2):
            range1 = arg1
            range2 = arg2
            # range1.start = range2.start
            # and range1.start included = range2.start included
            # and range1.end = range2.end
            # and range1.end included = range2.end included
            start1 = range1.getStart()
            end1 = range1.getEnd()
            start2 = range2.getStart()
            end2 = range2.getEnd()
            rc = self.resolveComparator(start1)
            return None if rc is None else self.BOOLEAN_TYPE.booleanAnd(
                rc.equalTo(start1, start2),
                range1.isStartIncluded() == range2.isStartIncluded(),
                rc.equalTo(end1, end2),
                range1.isEndIncluded() == range2.isEndIncluded()
            )
        else:
            return None

    @staticmethod
    def checkArguments(arg1: POINT_RANGE_UNION, arg2: POINT_RANGE_UNION) -> bool:
        return arg1 is None or arg2 is None

    def isPoint(self, arg1):
        return not isinstance(arg1, Range)

    def isRange(self, arg1):
        return isinstance(arg1, Range)

    def resolveComparator(self, arg):
        key = type(arg)
        cmp = COMPARATOR_MAP.get(key)
        return cmp



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use self file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
import logging
from typing import Set

from jdmn.runtime.DMNRuntimeException import DMNRuntimeException


class ModelElementRegistry:
    LOGGER = logging.getLogger(__name__)

    def __init__(self):
        self.map = {}

    def register(self, qName: str, className: str) -> None:
        if qName is None:
            raise DMNRuntimeException("Missing qName")
        if className is None:
            raise DMNRuntimeException("Missing class name")

        value = self.map[qName]
        if value is None:
            self.map[qName] = className
        elif value != className:
            raise DMNRuntimeException("Name '{}' is not unique".format(qName))
        else:
            self.LOGGER.warning("Name '%s' and value '%s' were already registered", qName, className)

    def discover(self, qName: str) -> str:
        return self.map[qName]

    def keys(self) -> Set[str]:
        return self.map.keys()



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use self file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
import datetime
from decimal import Decimal
from typing import Any
from unittest import TestCase

import isodate

from jdmn.runtime import DMNType
from jdmn.runtime.Context import Context
from jdmn.runtime.DMNRuntimeException import DMNRuntimeException


class Assert(TestCase):
    NUMBER_COMPARISON_PRECISION = 0.00000001

    def assertEquals(self, expected: Any, actual: Any, message: str = "") -> None:
        if expected is None:
            self.assertEqual(expected, actual, message)
        elif self.isNumber(expected):
            expectedBD = self.normalizeNumber(expected)
            actualBD = self.normalizeNumber(actual)
            if actual is None:
                self.assertEqual(expected, actual, message)
            else:
                condition = abs(expectedBD - actualBD) < self.NUMBER_COMPARISON_PRECISION
                self.assertTrue(condition, (message + ". Expected '{0}' found '{1}'").format(expectedBD, actualBD))
        elif self.isBoolean(expected):
            self.assertEqual(expected, actual, message)
        elif self.isString(expected):
            self.assertEqual(self.normalizeString(expected), self.normalizeString(actual), message)
        elif self.isDateTimeValue(expected):
            self.assertEqual(self.normalizeDateTime(expected), self.normalizeDateTime(actual), message)
        elif self.isList(expected):
            if actual is None:
                self.assertEqual(expected, actual, message)
            else:
                if message is None:
                    message = ""
                self.assertEqual(len(expected), len(actual), message + str(expected) + " vs " + str(actual))
                for i, exp in enumerate(expected):
                    self.assertEquals(exp, actual[i], message)
        elif self.isContext(expected):
            if actual is None:
                actual = Context()
            keySet = expected.getBindings().keys()
            for key in keySet:
                expectedMember = expected.get(key)
                actualMember = actual.get(key)
                self.assertEquals(expectedMember, actualMember, message + " for member '{0}'".format(key))
        elif self.isComplex(expected):
            for propertyName, expectedProperty in expected.__dict__.items():
                try:
                    actualProperty = getattr(actual, propertyName)
                    self.assertEquals(expectedProperty, actualProperty, message)
                except Exception as e:
                    raise DMNRuntimeException("Error in '{}.{}'".format(type(expected), str(propertyName)), e)
        else:
            self.assertEqual(expected, actual, message)

    def isNumber(self, obj: Any) -> bool:
        return isinstance(obj, Decimal)

    def isString(self, obj: Any) -> bool:
        return isinstance(obj, str)

    def isBoolean(self, obj: Any) -> bool:
        return isinstance(obj, bool)

    def isDate(self, obj: Any) -> bool:
        return isinstance(obj, datetime.date)

    def isTime(self, obj: Any) -> bool:
        return isinstance(obj, datetime.time)

    def isDateTime(self, obj: Any) -> bool:
        return isinstance(obj, datetime.datetime)

    def isDuration(self, obj: Any) -> bool:
        return isinstance(obj, isodate.Duration) or isinstance(obj, datetime.timedelta)

    def isDateTimeValue(self, obj: Any) -> bool:
        return self.isDate(obj) \
               or self.isTime(obj) \
               or self.isDateTime(obj) \
               or self.isDuration(obj)

    def isList(self, obj: Any) -> bool:
        return isinstance(obj, list)

    def isContext(self, obj: Any) -> bool:
        return isinstance(obj, Context)

    def isComplex(self, obj: Any) -> bool:
        return isinstance(obj, DMNType.DMNType)

    def normalizeDateTime(self, obj: Any) -> Any:
        if obj is None:
            return None
        return obj

    def normalizeNumber(self, obj: Any) -> Any:
        if obj is None:
            return None
        if isinstance(obj, str):
            return float(obj)
        else:
            return obj

    @staticmethod
    def normalizeString(string):
        # Replace surrogate pairs with equivalent
        if string is None:
            return None

        # Roundtrip to utf-16 to replace surrogate pairs
        utf16bytes = string.encode('utf-16', 'surrogatepass')
        utf16chars = utf16bytes.decode('utf-16')
        utf8 = utf16chars.encode('utf-8').decode('utf-8')
        return utf8



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use self file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
import decimal
from decimal import Decimal

from isodate import Duration

from jdmn.feel.lib.Types import DURATION, BOOLEAN, DECIMAL
from jdmn.feel.lib.type.time.DefaultCalendarType import DefaultCalendarType
from jdmn.feel.lib.type.time.DefaultDurationComparator import DefaultDurationComparator
from jdmn.runtime.DMNRuntimeException import DMNRuntimeException


class DefaultDurationType(DefaultCalendarType):
    def __init__(self):
        super().__init__()
        self.__durationComparator = DefaultDurationComparator()

    def durationIs(self, first: DURATION, second: DURATION) -> BOOLEAN:
        if first is None or second is None:
            return first == second

        if self.isYearsAndMonthsDuration(first) and self.isYearsAndMonthsDuration(second):
            return first == second
        elif self.isDaysAndTimeDuration(first) and self.isDaysAndTimeDuration(second):
            return first == second
        else:
            return False

    def durationEqual(self, first: DURATION, second: DURATION) -> BOOLEAN:
        return self.__durationComparator.equalTo(first, second)

    def durationNotEqual(self, first: DURATION, second: DURATION) -> BOOLEAN:
        return self.__durationComparator.notEqualTo(first, second)

    def durationLessThan(self, first: DURATION, second: DURATION) -> BOOLEAN:
        return self.__durationComparator.lessThan(first, second)

    def durationGreaterThan(self, first: DURATION, second: DURATION) -> BOOLEAN:
        return self.__durationComparator.greaterThan(first, second)

    def durationLessEqualThan(self, first: DURATION, second: DURATION) -> BOOLEAN:
        return self.__durationComparator.lessEqualThan(first, second)

    def durationGreaterEqualThan(self, first: DURATION, second: DURATION) -> BOOLEAN:
        return self.__durationComparator.greaterEqualThan(first, second)

    def durationAdd(self, first: DURATION, second: DURATION) -> DURATION:
        if first is None or second is None:
            return None

        if self.isYearsAndMonthsDuration(first) and self.isYearsAndMonthsDuration(second):
            totalMonths = self.monthsValue(first) + self.monthsValue(second)
            return Duration(months=totalMonths)
        elif self.isDaysAndTimeDuration(first) and self.isDaysAndTimeDuration(second):
            durationInSeconds = self.secondsValue(first) + self.secondsValue(second)
            return Duration(seconds=durationInSeconds)
        else:
            raise DMNRuntimeException("Cannot add '{}' and '{}'".format(first, second))

    def durationSubtract(self, first: DURATION, second: DURATION) -> DURATION:
        if first is None or second is None:
            return None

        return self.durationAdd(first, second.__neg__())

    def durationDivide(self, first: DURATION, second: DURATION) -> DECIMAL:
        if first is None or second is None:
            return None

        if self.isYearsAndMonthsDuration(first) and self.isYearsAndMonthsDuration(second):
            firstValue = Decimal(self.monthsValue(first))
            secondValue = Decimal(self.monthsValue(second))
            return self.divideNumber(firstValue, secondValue)
        elif self.isDaysAndTimeDuration(first) and self.isDaysAndTimeDuration(second):
            firstValue = Decimal(self.secondsValue(first))
            secondValue = Decimal(self.secondsValue(second))
            return self.divideNumber(firstValue, secondValue)
        else:
            raise DMNRuntimeException("Cannot divide '{}' and '{}'".format(first, second))

    def durationMultiplyNumber(self, first: DURATION, second: DECIMAL) -> DURATION:
        if first is None or second is None:
            return None

        if self.isYearsAndMonthsDuration(first):
            firstValue = Decimal(self.monthsValue(first))
            totalMonths = self.multiplyNumber(firstValue, second)
            return Duration(months=totalMonths.to_integral())
        elif self.isDaysAndTimeDuration(first):
            firstValue = Decimal(self.secondsValue(first))
            durationInSeconds = int(self.multiplyNumber(firstValue, second))
            return Duration(seconds=durationInSeconds)
        else:
            raise DMNRuntimeException("Cannot multiply '{}' and '{}'".format(first, second))

    def durationDivideNumber(self, first: DURATION, second: DECIMAL) -> DURATION:
        if first is None or second is None:
            return None

        if self.isYearsAndMonthsDuration(first):
            firstValue = Decimal(self.monthsValue(first))
            totalMonths = self.divideNumber(firstValue, second)
            return Duration(months=totalMonths.to_integral())
        elif self.isDaysAndTimeDuration(first):
            firstValue = Decimal(self.secondsValue(first))
            durationInSeconds = int(self.divideNumber(firstValue, second))
            return Duration(seconds=durationInSeconds)
        else:
            raise DMNRuntimeException("Cannot divide '{}' and '{}'".format(first, second))

    @staticmethod
    def multiplyNumber(firstValue: DECIMAL, secondValue: DECIMAL) -> DECIMAL:
        # DECIMAL 128
        with decimal.localcontext() as ctx:
            ctx.prec = 34
            return firstValue * secondValue

    @staticmethod
    def divideNumber(firstValue: DECIMAL, secondValue: DECIMAL) -> DECIMAL:
        if secondValue == 0:
            return None

        # DECIMAL 128
        with decimal.localcontext() as ctx:
            ctx.rounding = decimal.ROUND_HALF_DOWN
            ctx.prec = 34
            return firstValue / secondValue
